# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from six import itervalues

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.asm import Asm
from f5.sdk_exception import UnsupportedTmosVersion


AGENT_LB_DEL_ORDER = {'/mgmt/tm/ltm/virtual': 1,
                      '/mgmt/tm/ltm/pool': 2,
                      'mgmt/tm/ltm/node/': 3,
                      'monitor': 4,
                      'virtual-address': 5,
                      'mgmt/tm/net/self/': 6,
                      '/mgmt/tm/net/fdb': 7,
                      'mgmt/tm/net/tunnels/tunnel/': 8,
                      'mgmt/tm/net/tunnels/vxlan/': 9,
                      'mgmt/tm/net/tunnels/gre': 10,
                      'mgmt/tm/net/vlan': 11,
                      'route': 12,
                      '/mgmt/tm/ltm/snatpool': 13,
                      '/mgmt/tm/ltm/snat-translation': 14,
                      '/mgmt/tm/net/route-domain': 15,
                      '/mgmt/tm/sys/folder': 16}


def order_by_weights(unordered, weights_table):
    '''(iterable, wieghts) --> iterable_ordered_by_weights

    AGENT_LB_DEL_ORDER is an example weights table.  Pass it as the second
    argument where the first is a list of BigIP device URIs and it will return
    a list with the resources in the appropriate order for deletion.
    Note that URIS that do not match a key in the weights table are set to have
    the same "high" weight.   Their relative order will not change.

    >>> order_by_weights([URI1, URI2, ...], AGENT_LB_DEL_ORDER)
    [URI2, URI1, ....]
    '''
    min_minus_one = min(weights_table.values()) - 1

    def order_key(item):
        for k in weights_table:
            if k in item:
                return weights_table[k]
        return min_minus_one
    ordered_by_weights = sorted(list(unordered), key=order_key)
    return ordered_by_weights


def register_collection_atoms(collection):
    '''Given a collection return a registry of all of its atoms (elements).

    Registries are dictionaries with selfLink keys, and PathElement values.
    A registry provides a snapshot of all resources of a type on the device.
    '''
    resource_registry = {}
    try:
        resources = collection.get_collection()
    except Exception as ex:
        logging.debug(ex)
        return resource_registry
    for resource in resources:
        try:
            resource_registry[resource.selfLink] = resource
        except KeyError as ex:
            raise ex
        except UnsupportedTmosVersion as ex:
            logging.debug(ex)
            continue
    return resource_registry


def register_OC_atoms(organizing_collection):
    '''Given an OrganizingCollection (OC) return a registry of its atoms.

    Registries are dictionaries with selfLink keys, and PathElement values.
    A registry provides a snapshot of all resources of a type on the device.
    '''
    OC_atoms_registry = {}
    OC_types = organizing_collection._meta_data['allowed_lazy_attributes']
    # Removing ASM as this causes some intermittent problems on Jenkins,
    # whereby device snapshot tries to delete factory signatures, resulting in
    # REST error and test failures when
    if Asm in OC_types:
        OC_types.remove(Asm)
    for OC_type in OC_types:
        try:
            lazy_instance =\
                getattr(organizing_collection, OC_type.__name__.lower())
        except UnsupportedTmosVersion as ex:
            logging.debug(ex)
            continue
        if isinstance(lazy_instance, Collection):
            OC_atoms_registry.update(register_collection_atoms(lazy_instance))
        elif isinstance(lazy_instance, OrganizingCollection):
            OC_atoms_registry.update(register_OC_atoms(lazy_instance))
    return OC_atoms_registry


def register_device(mgmt_rt):
    OCs = [getattr(mgmt_rt, c.__name__.lower())
           for c in mgmt_rt._meta_data['allowed_lazy_attributes']]
    grand_registry = {}
    for OC in OCs:
        grand_registry.update(register_OC_atoms(OC))
    return grand_registry


def register_loadbalancer_elements(mgmt_rt):
    monitor_registry = register_OC_atoms(mgmt_rt.tm.ltm.monitor)
    pool_registry = register_collection_atoms(mgmt_rt.tm.ltm.pools)
    snat_registry = register_collection_atoms(mgmt_rt.tm.ltm.snats)
    virtual_registry = register_collection_atoms(mgmt_rt.tm.ltm.virtuals)
    virtual_address_s_registry =\
        register_collection_atoms(mgmt_rt.tm.ltm.virtual_address_s)
    member_registry = {}
    for pool in list(itervalues(pool_registry)):
        mc = pool.members_s
        member_registry.update(register_collection_atoms(mc))
    folder_registry = register_collection_atoms(mgmt_rt.tm.sys.folders)
    registries = {'monitor_registry': monitor_registry,
                  'pool_registry': pool_registry,
                  'snat_registry': snat_registry,
                  'virtual_registry': virtual_registry,
                  'virtual_address_s_registry': virtual_address_s_registry,
                  'member_registry': member_registry,
                  'folder_registry': folder_registry}
    return registries



# coding=utf-8
#
#  Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®) Topology Records module.

REST URI
    ``http://localhost/mgmt/tm/gtm/listener``

GUI Path
    ``DNS --> Delivery : Topology : Records``

REST Kind
    ``tm:gtm:listener:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedOperation
from icontrol.exceptions import iControlUnexpectedHTTPError


class Listeners(Collection):
    """BIG-IP® GTM Listener collection"""
    def __init__(self, gtm):
        super(Listeners, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Listener]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:listener:listenerstate': Listener}


class Listener(Resource):
    """BIG-IP® GTM Listener resource"""
    def __init__(self, listeners):
        super(Listener, self).__init__(listeners)
        self._meta_data['required_json_kind'] = 'tm:gtm:listener:listenerstate'
        self._meta_data['required_creation_parameters'].update(('address',))
        self._meta_data['attribute_registry'] = {
            'tm:gtm:listener:profiles:profilescollectionstate':
                Profiles_s
        }

    def exists(self, **kwargs):
        # This endpoint does not return status 404 if the GET request
        # does not find the resource. On one installation it returned
        # 400, and on another it returned 500
        try:
            result = super(Listener, self).exists(**kwargs)
        except iControlUnexpectedHTTPError as ex:
            if 'listener does not exist' in str(ex):
                return False
            else:
                raise
        return result


class Profiles_s(Collection):
    """BIG-IP® GTM Listener Profile sub-collection"""
    def __init__(self, server):
        super(Profiles_s, self).__init__(server)
        self._meta_data['allowed_lazy_attributes'] = [Profile]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:listener:profiles:profilescollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:listener:profiles:profilesstate':
                Profile}


class Profile(Resource):
    """BIG-IP® GTM Listener Profile sub-collection

    Since GTM listener is a wrapper for LTM virtual,
    profile removal and profile attachment should be done via the created
    LTM virtual. Only loading or refresh of profiles is supported.
    Remainder of operations(Create, Update, Modify, Delete) should be done via
    LTM Virtual Server Profiles endpoint.
    """

    def __init__(self, profiles):
        super(Profile, self).__init__(profiles)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:listener:profiles:profilesstate'
        self._meta_data['required_load_parameters'].update(('partition',))

    def create(self, **kwargs):
        """Create is not supported for profile sub-collection

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )

    def modify(self, **kwargs):
        """Modify is not supported for profile sub-collection

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )

    def update(self, **kwargs):
        """Update is not supported for profile sub-collection

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the update method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for profile sub-collection

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager™ (LTM®) LSN pool module.

REST URI
    ``http://localhost/mgmt/tm/ltm/lsn-pool``
    ``http://localhost/mgmt/tm/ltm/lsn-log-profile``

GUI Path
    ``Carrier Grade NAT --> LSN Pools``
    ``Carrier Grade NAT --> Logging Profiles --> LSN``

REST Kind
    ``tm:ltm:lsn-pool:*``
    ``tm:ltm:lsn-log-profile:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class LSN_Pools(Collection):
    """BIG-IP® LTM LSN pool collection"""
    def __init__(self, ltm):
        super(LSN_Pools, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [LSN_Pool]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:lsn-pool:lsn-poolstate': LSN_Pool}


class LSN_Pool(Resource):
    """BIG-IP® LTM LSN pool resource"""
    def __init__(self, lsnpool_s):
        super(LSN_Pool, self).__init__(lsnpool_s)
        self._meta_data['required_json_kind'] = 'tm:ltm:lsn-pool:lsn-poolstate'
        self._meta_data['allowed_lazy_attributes'] = [LSN_Log_Profile]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:lsn-log-profile:lsn-log-profilestate': LSN_Log_Profile}


class LSN_Log_Profiles(Collection):
    """BIG-IP® LTM LSN pool log profile collection"""
    def __init__(self, profile):
        super(LSN_Log_Profiles, self).__init__(profile)
        self._meta_data['allowed_lazy_attributes'] = [LSN_Log_Profile]
        self._meta_data['required_json_kind'] = (
            'tm:ltm:lsn-log-profile:lsn-log-profilecollectionstate')
        self._meta_data['attribute_registry'] = \
            {'tm:ltm:lsn-log-profile:lsn-log-profilestate': LSN_Log_Profile}
        self._meta_data['minimum_version'] = '11.6.0'


class LSN_Log_Profile(Resource):
    """BIG-IP® LTM LSN pool log profile resource"""
    def __init__(self, LSNLogProfile_s):
        super(LSN_Log_Profile, self).__init__(LSNLogProfile_s)
        self._meta_data['required_json_kind'] = (
            'tm:ltm:lsn-log-profile:lsn-log-profilestate')



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager (LTM) node module.

REST URI
    ``http://localhost/mgmt/tm/ltm/node``

GUI Path
    ``Local Traffic --> Nodes``

REST Kind
    ``tm:ltm:node:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import NodeStateModifyUnsupported


class Nodes(Collection):
    """BIG-IP® LTM node collection"""
    def __init__(self, ltm):
        super(Nodes, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Node]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:node:nodestate': Node}


class Node(Resource):
    """BIG-IP® LTM node resource"""
    def __init__(self, nodes):
        super(Node, self).__init__(nodes)
        self._meta_data['required_json_kind'] = 'tm:ltm:node:nodestate'
        self._meta_data['required_creation_parameters'].update(
            ('partition', 'address',)
        )
        self._meta_data['read_only_attributes'].append('ephemeral')
        self._meta_data['read_only_attributes'].append('address')

    def update(self, **kwargs):
        """Call this to change the configuration of the service on the device.

        This method uses HTTP PUT alter the service state on the device.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.  Various edge cases are handled:

        * read-only attributes that are unchangeable are removed
        * If ``fqdn`` is in the kwargs or set as an attribute, removes the
          ``autopopulate`` and ``addressFamily`` keys from it if there.

        :param kwargs: keys and associated values to alter on the device

        """
        checked = self._check_node_parameters(**kwargs)
        return self._update(**checked)

    def _check_node_parameters(self, **kwargs):
        """See discussion in issue #840."""
        if 'fqdn' in kwargs:
            kwargs['fqdn'].pop('autopopulate', '')
            kwargs['fqdn'].pop('addressFamily', '')
        if 'fqdn' in self.__dict__:
            self.__dict__['fqdn'].pop('autopopulate', '')
            self.__dict__['fqdn'].pop('addressFamily', '')
        if 'state' in kwargs:
            if kwargs['state'] != 'user-up' and kwargs['state'] != \
                    'user-down':
                kwargs.pop('state')
        if 'state' in self.__dict__:
            if self.__dict__['state'] != 'user-up' and self.__dict__['state'] \
                    != 'user-down':
                self.__dict__.pop('state')
        if 'session' in kwargs:
            if kwargs['session'] != 'user-enabled' and kwargs['session'] != \
                    'user-disabled':
                kwargs.pop('session')
        if 'session' in self.__dict__:
            if self.__dict__['session'] != 'user-enabled' and \
                    self.__dict__['session'] != 'user-disabled':
                self.__dict__.pop('session')
        # Until we implement sanity checks for __dict__ this needs to stay here
        self.__dict__.pop('ephemeral', '')
        self.__dict__.pop('address', '')
        return kwargs

    def _modify(self, **patch):
        """Override modify to check kwargs before request sent to device."""
        if 'state' in patch:
            if patch['state'] not in ['user-up', 'user-down', 'unchecked', 'fqdn-up']:
                msg = "The node resource does not support a modify with the " \
                      "value of the 'state' attribute as %s. The accepted " \
                      "values are 'user-up', 'user-down', 'unchecked', or 'fqdn-up'" \
                      % patch['state']
                raise NodeStateModifyUnsupported(msg)
        if 'session' in patch:
            if patch['session'] not in ['user-enabled', 'user-disabled']:
                msg = "The node resource does not support a modify with the " \
                      "value of the 'session' attribute as %s. " \
                      "The accepted values are 'user-enabled' or " \
                      "'user-disabled'" % patch['session']
                raise NodeStateModifyUnsupported(msg)
        super(Node, self)._modify(**patch)



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Application Security Manager™ (ASM®) module.

REST URI
    ``http://localhost/mgmt/tm/asm/``

GUI Path
    ``Security``

REST Kind
    ``tm:asm:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.asm.attack_types import Attack_Types_s
from f5.bigip.tm.asm.file_transfer import File_Transfer
from f5.bigip.tm.asm.policies import Policies_s
from f5.bigip.tm.asm.policy_templates import Policy_Templates_s
from f5.bigip.tm.asm.signature_sets import Signature_Sets_s
from f5.bigip.tm.asm.signature_statuses import Signature_Statuses_s
from f5.bigip.tm.asm.signature_systems import Signature_Systems_s
from f5.bigip.tm.asm.signature_update import Signature_Update
from f5.bigip.tm.asm.signatures import Signatures_s
from f5.bigip.tm.asm.tasks import Tasks


class Asm(OrganizingCollection):
    """BIG-IP® Application Security Manager (ASM) organizing

    collection.
    """

    def __init__(self, tm):
        super(Asm, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Attack_Types_s,
            File_Transfer,
            Policies_s,
            Policy_Templates_s,
            Signature_Sets_s,
            Signature_Statuses_s,
            Signature_Systems_s,
            Signature_Update,
            Signatures_s,
            Tasks,
        ]



# coding=utf-8
#
"""Classes and functions for configuring BIG-IP"""
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from f5.bigip.mixins import FileDownloadMixin
from f5.bigip.mixins import FileUploadMixin
from f5.bigip.resource import PathElement
from f5.sdk_exception import FileMustNotHaveDotISOExtension


class File_Transfer(PathElement):
    """A PathElement for File_Transfer resources."""
    def __init__(self, shared):
        super(File_Transfer, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Bulk,
            Madm,
            Uploads,
            Ucs_Uploads,
            Ucs_Downloads
        ]


class Uploads(PathElement, FileUploadMixin):
    """A file upload resource."""
    def __init__(self, file_transfer):
        super(Uploads, self).__init__(file_transfer)

    def upload_file(self, filepathname, **kwargs):
        filename = os.path.basename(filepathname)
        if os.path.splitext(filename)[-1] == '.iso':
            raise FileMustNotHaveDotISOExtension(filename)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._upload_file(filepathname, **kwargs)

    def upload_stringio(self, stringio, target, **kwargs):
        self.file_bound_uri = self._meta_data['uri'] + target
        self._upload(stringio, **kwargs)

    def upload_bytes(self, bytestring, target, **kwargs):
        self.file_bound_uri = self._meta_data['uri'] + target
        self._upload(StringIO(bytestring), **kwargs)


class Bulk(PathElement, FileUploadMixin, FileDownloadMixin):
    """A file upload resource."""
    def __init__(self, file_transfer):
        super(Bulk, self).__init__(file_transfer)
        self._meta_data['minimum_version'] = '13.0.0'

    def download_file(self, src, dest, **kwargs):
        filename = os.path.basename(src)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._download_file(src, dest, **kwargs)


class Madm(PathElement, FileDownloadMixin):
    """A file upload resource."""
    def __init__(self, file_transfer):
        super(Madm, self).__init__(file_transfer)

    def download_file(self, src, dest, **kwargs):
        filename = os.path.basename(src)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._download_file(src, dest, **kwargs)


class Ucs_Uploads(PathElement, FileUploadMixin):
    """A ucs file upload resource"""
    def __init__(self, file_transfer):
        super(Ucs_Uploads, self).__init__(file_transfer)

    def upload_file(self, filepathname, **kwargs):
        filename = os.path.basename(filepathname)
        if os.path.splitext(filename)[-1] == '.iso':
            raise FileMustNotHaveDotISOExtension(filename)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._upload_file(filepathname, **kwargs)

    def upload_bytes(self, bytestring, target, **kwargs):
        self.file_bound_uri = self._meta_data['uri'] + target
        self._upload(StringIO(bytestring), **kwargs)


class Ucs_Downloads(PathElement, FileDownloadMixin):
    """A ucs file download resource"""
    def __init__(self, file_transfer):
        super(Ucs_Downloads, self).__init__(file_transfer)

    def download_file(self, src, dest, **kwargs):
        filename = os.path.basename(src)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._download_file(src, dest, **kwargs)



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from f5.sdk_exception import UnsupportedOperation
from icontrol.exceptions import iControlUnexpectedHTTPError


class Session_Tracking_Statuses_s(Collection):
    """BIG-IP® ASM Session Tracking Statuses sub-collection."""
    def __init__(self, policy):
        super(Session_Tracking_Statuses_s, self).__init__(policy)
        self._meta_data['object_has_stats'] = False
        self._meta_data['minimum_version'] = '11.6.0'
        self._meta_data['allowed_lazy_attributes'] = [Session_Tracking_Status]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:session-tracking-statuses:session-tracking-statuscollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:session-tracking-statuses:session-tracking-statusstate': Session_Tracking_Status
        }


class Session_Tracking_Status(AsmResource):
    """BIG-IP® ASM Session TrackingStatus resource."""
    def __init__(self, session_tracking_statuses_s):
        super(Session_Tracking_Status, self).__init__(
            session_tracking_statuses_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:session-tracking-statuses:session-tracking-statusstate'
        self._meta_data['required_creation_parameters'] = {
            'action', 'scope', 'value'
        }

    def modify(self, **kwargs):
        """Modify is not supported for Session Tracking resource

        :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )

    def refresh(self, **kwargs):
        # If delete() was called, then there is no reason to check if something
        # exists in the loop below.
        deleted = False
        if 'deleted' in self.__dict__ and self.__dict__['deleted'] is True:
            deleted = True
        if deleted:
            return self._refresh(**kwargs)

        # It is possible that even after creation, ASM will not correctly be
        # reporting creation of a Session Tracking Status. Therefore, try several
        # times before we finally fail.
        ex = iControlUnexpectedHTTPError(
            "Failed to refresh the session-tracking-status"
        )
        for _ in range(0, 30):
            try:
                return self._refresh(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def load(self, **kwargs):
        # If delete() was called, then there is no reason to check if something
        # exists in the loop below.
        deleted = False
        if 'deleted' in self.__dict__ and self.__dict__['deleted'] is True:
            deleted = True
        if deleted:
            return self._load(**kwargs)

        # It is possible that even after creation, ASM will not correctly be
        # reporting creation of a Session Tracking Status. Therefore, try several
        # times before we finally fail.
        ex = iControlUnexpectedHTTPError(
            "Failed to load the session-tracking-status"
        )
        for _ in range(0, 30):
            try:
                return self._load(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def delete(self, **kwargs):
        # If delete() was called, then there is no reason to check if something
        # exists in the loop below.
        deleted = False
        if 'deleted' in self.__dict__ and self.__dict__['deleted'] is True:
            deleted = True
        if deleted:
            return self._delete(**kwargs)

        # It is possible that even after creation, ASM will not correctly be
        # reporting creation of a Session Tracking Status. Therefore, try several
        # times before we finally fail.
        ex = iControlUnexpectedHTTPError(
            "Failed to delete the session-tracking-status"
        )
        for _ in range(0, 30):
            try:
                return self._delete(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise
        raise ex

    def _check_exception(self, ex):
        retryable = [
            # Retry potential failure to load due to high policy create/delete
            #
            # There is a possibility that due to a large number of successive
            # create/delete actions on the policies endpoint, that the collection
            # known to the REST API is changing repeatedly.
            #
            # A consequence of this is that the load method will invariably fail
            # intermittently with this error
            #
            # {
            #   "code":404,
            #   "message": "Could not get the Session Awareness Data Point
            #              \'Block All for Username fake2\', No matching record was found.",
            #   "referer":"10.0.2.2",
            #   "restOperationId":9272,
            #   "errorStack":[
            #     "ASMConfigException(error_message:Could not get the Session Awareness Data
            #      Point \'Block All for Username fake2\', No matching record was found.,
            #      error_code:NONEXISTENT_RECORD, internal_error:Failed get on nonexistent
            #      record for SessionAwarenessDataPoint found -- get aborted,
            #      rest_code:REST_NOT_FOUND)"
            #
            'Could not get the Session Awareness Data Point',

            'Could not delete the Session Awareness Data Point'
        ]
        if any(x in str(ex) for x in retryable):
            time.sleep(3)
            return True
        elif 'errorStack' in ex:
            stack = ' '.join(ex['errorStack'])
            if any(x in stack for x in retryable):
                time.sleep(3)
                return True
            else:
                return False
        else:
            return False



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""This provides tools that manipulate configurations of device resources.


"""

from six import iteritems
from six.moves.urllib.parse import urlparse


class BaseHandler(object):
    def __init__(self, resource):
        self.resource = resource


class Stats(BaseHandler):
    """Stats handler class

    When this class is first instantiated, it will populate, ['stat']
    attribute with individual stats, that can be accessed via dot notation
    """
    def __init__(self, stats):
        super(Stats, self).__init__(stats)
        self.rdict = self.resource.entries
        self._update_stats()

    def _key_dot_replace(self, rdict):
        """Replace fullstops in returned keynames"""
        temp_dict = {}
        for key, value in iteritems(rdict):
            if isinstance(value, dict):
                value = self._key_dot_replace(value)
            temp_dict[key.replace('.', '_')] = value
        return temp_dict

    def _get_nest_stats(self):
        """Helper method to deal with nestedStats

        as json format changed in v12.x
        """
        for x in self.rdict:
            check = urlparse(x)
            if check.scheme:
                nested_dict = self.rdict[x]['nestedStats']
                tmp_dict = nested_dict['entries']
                return self._key_dot_replace(tmp_dict)

        return self._key_dot_replace(self.rdict)

    def _update_stats(self):
        """Attaches stat attribute to stats object"""
        stat_vals = self._get_nest_stats()
        self.__dict__['stat'] = DottedDict(stat_vals)

    def refresh(self, **kwargs):
        """Refreshes stats attached to an object"""
        self.resource.refresh(**kwargs)
        self.rdict = self.resource.entries
        self._update_stats()


class DottedDict(dict):
    """Helper class

    Allows accessing contents of a given dictionary
    by using object like 'dot' notation.

    @param :rdict Dictionary
    """
    def __init__(self, rdict):
        super(DottedDict, self).__init__()
        self.update(rdict)

    def __getattr__(self, k):
        if isinstance(self[k], dict) and not isinstance(self[k], DottedDict):
            self[k] = DottedDict(self[k])
        return self[k]



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Guest (vcmp) module

REST URI
    ``http://localhost/mgmt/tm/vcmp/virtual-disk/``

GUI Path
    ``Virtual Disk List``

REST Kind
    ``tm:vcmp:virtual-disk:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import DisallowedCreationParameter
from f5.sdk_exception import DisallowedReadParameter
from f5.sdk_exception import F5SDKError
from f5.sdk_exception import UnsupportedMethod


class Virtual_Disks(Collection):
    def __init__(self, vcmp):
        super(Virtual_Disks, self).__init__(vcmp)
        self._meta_data['allowed_lazy_attributes'] = [Virtual_Disk]
        self._meta_data['required_json_kind'] = 'tm:vcmp:virtual-disk:virtual-diskcollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:vcmp:virtual-disk:virtual-diskstate': Virtual_Disk
        }


class Virtual_Disk(Resource):
    def __init__(self, collection):
        super(Virtual_Disk, self).__init__(collection)
        self._meta_data['required_json_kind'] = 'tm:vcmp:virtual-disk:virtual-diskstate'

    def load(self, **kwargs):
        """Loads a given resource

        Loads a given resource provided a 'name' and an optional 'slot'
        parameter. The 'slot' parameter is not a required load parameter
        because it is provided as an optional way of constructing the
        correct 'name' of the vCMP resource.

        :param kwargs:
        :return:
        """
        kwargs['transform_name'] = True
        kwargs = self._mutate_name(kwargs)
        return self._load(**kwargs)

    def exists(self, **kwargs):
        kwargs['transform_name'] = True
        kwargs = self._mutate_name(kwargs)
        return self._exists(**kwargs)

    def delete(self, **kwargs):
        kwargs['transform_name'] = True
        kwargs = self._mutate_name(kwargs)
        return self._delete(**kwargs)

    def modify(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the modify method" % self.__class__.__name__
        )

    def create(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the create method" % self.__class__.__name__
        )

    def update(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )

    def _check_load_parameters(self, **kwargs):
        """Override method for one in resource.py to check partition

        The partition cannot be included as a parameter to load a guest.
        Raise an exception if a consumer gives the partition parameter.

        :raises: DisallowedReadParameter
        """

        if 'partition' in kwargs:
            msg = "'partition' is not allowed as a load parameter. Vcmp " \
                "guests are accessed by name."
            raise DisallowedReadParameter(msg)
        super(Virtual_Disk, self)._check_load_parameters(**kwargs)

    def _check_create_parameters(self, **kwargs):
        """Override method for one in resource.py to check partition

        The partition cannot be included as a parameter to create a guest.
        Raise an exception if a consumer gives the partition parameter.

        :raises: DisallowedCreationParameter
        """

        if 'partition' in kwargs:
            msg = "'partition' is not allowed as a create parameter. Vcmp " \
                "guests are created with the 'name' at least."
            raise DisallowedCreationParameter(msg)
        super(Virtual_Disk, self)._check_create_parameters(**kwargs)

    def _mutate_name(self, kwargs):
        slot = kwargs.pop('slot', None)
        if slot is not None:
            try:
                kwargs['name'] = '{0}/{1}'.format(kwargs['name'], int(slot))
            except ValueError:
                raise F5SDKError(
                    "The provided 'slot' must be a number"
                )
        return kwargs



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® system folder module

REST URI
    ``http://localhost/mgmt/tm/sys/folder``

REST Kind
    ``tm:sys:folder:*``
"""
import logging
from requests.exceptions import HTTPError

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource

logger = logging.getLogger(__name__)


class Folders(Collection):
    """BIG-IP® system folder collection.

    These are what we refer to as ``partition`` in the SDK.
    """
    def __init__(self, sys):
        super(Folders, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [Folder]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:folder:folderstate': Folder}


class Folder(Resource):
    def __init__(self, folder_s):
        '''BIG-IP® system folder resource.

        Folder objects are the same as the partition so we need to deal with
        them slightly differently than other Resources.  For example when
        you refresh/load them you need to use the partition name instead of
        the partition and name because their self link looks something like
        `https://localhost/mgmt/tm/sys/folder/~testfolder`.  Notice that there
        is no ~partition~name format for the object.
        '''
        super(Folder, self).__init__(folder_s)
        self._meta_data['required_json_kind'] = 'tm:sys:folder:folderstate'
        # refresh() and load() require partition, not name
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_creation_parameters'].update(('subPath',))

    def _create_subpath_uri(self, kwargs):
        base_uri = self._meta_data['container']._meta_data['uri']
        name = kwargs.pop('name', '')
        partition = kwargs.pop('partition', '')
        if not name and not partition:
            # Root folder - https://localhost/mgmt/tm/sys/folder/~
            load_uri = base_uri + '~'
        elif not name and partition:
            # Top level - https://localhost/mgmt/tm/sys/folder/~partition
            load_uri = base_uri + '~' + partition
        elif name and not partition:
            # Top level - https://localhost/mgmt/tm/sys/folder/~partition
            load_uri = base_uri + '~' + name
        else:
            # Nested Folder (allow for name to be many folders)
            # https://localhost/mgmt/tm/sys/folder/~partition~f1~f2
            name = name.replace('/', '~')
            load_uri = base_uri + '~' + partition + '~' + name
        return load_uri

    def _load(self, **kwargs):
        read_session = self._meta_data['bigip']._meta_data['icr_session']
        load_uri = self._create_subpath_uri(kwargs)
        response = read_session.get(load_uri, uri_as_parts=False, **kwargs)
        return self._produce_instance(response)

    def update(self, **kwargs):
        '''Update the object, removing device group if inherited

        If inheritedDevicegroup is the string "true" we need to remove
        deviceGroup from the args before we update or we get the
        following error:

        The floating traffic-group: /Common/traffic-group-1 can only be set on
        /testfolder if its device-group is inherited from the root folder
        '''
        inherit_device_group = self.__dict__.get('inheritedDevicegroup', False)
        if inherit_device_group == 'true':
            self.__dict__.pop('deviceGroup')
        return self._update(**kwargs)

    def exists(self, **kwargs):
        """Check for the existence of the named object on the BIG-IP

        Tries to `load()` the object and if it fails checks the exception
        for 404.  If the `load()` is successful it returns `True` if the
        exception is :exc:`requests.HTTPError` and the
        ``status_code`` is ``404``
        it will return error.  All other errors are raised as is.

        :param kwargs: Keyword arguments required to get objects
        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: bool -- The objects exists on BIG-IP® or not.
        :raises: :exc:`requests.HTTPError`, Any HTTP error that was not status
            code 404.
        """

        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        load_uri = self._create_subpath_uri(kwargs)
        session = self._meta_data['bigip']._meta_data['icr_session']
        kwargs.update(requests_params)
        try:
            session.get(load_uri, **kwargs)
        except HTTPError as err:
            logger.debug(err.response.text)
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network tunnels module.

REST URI
    ``http://localhost/mgmt/tm/net/tunnels``

GUI Path
    ``Network --> tunnels``

REST Kind
    ``tm:net:tunnels:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class TunnelS(OrganizingCollection):
    """BIG-IP® network tunnels collection"""
    def __init__(self, net):
        super(TunnelS, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [
            Gres,
            Tunnels,
            Vxlans,
        ]


class Tunnels(Collection):
    """BIG-IP® network tunnels resource (collection for GRE, Tunnel, VXLANs"""
    def __init__(self, tunnelS):
        super(Tunnels, self).__init__(tunnelS)
        self._meta_data['allowed_lazy_attributes'] = [Gres, Tunnel, Vxlans]
        self._meta_data['attribute_registry'] =\
            {'tm:net:tunnels:tunnel:tunnelstate': Tunnel}


class Tunnel(Resource):
    """BIG-IP® tunnels tunnel resource"""
    def __init__(self, tunnels):
        super(Tunnel, self).__init__(tunnels)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:net:tunnels:tunnel:tunnelstate'


class Gres(Collection):
    """BIG-IP® tunnels GRE sub-collection"""
    def __init__(self, tunnels):
        super(Gres, self).__init__(tunnels)
        self._meta_data['allowed_lazy_attributes'] = [Gre]
        self._meta_data['attribute_registry'] =\
            {'tm:net:tunnels:gre:grestate': Gre}


class Gre(Resource):
    """BIG-IP® tunnels GRE sub-collection resource"""
    def __init__(self, gres):
        super(Gre, self).__init__(gres)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:net:tunnels:gre:grestate'


class Vxlans(Collection):
    """BIG-IP® tunnels VXLAN sub-collection"""
    def __init__(self, tunnels):
        super(Vxlans, self).__init__(tunnels)
        self._meta_data['allowed_lazy_attributes'] = [Vxlan]
        self._meta_data['attribute_registry'] =\
            {'tm:net:tunnels:vxlan:vxlanstate': Vxlan}


class Vxlan(Resource):
    """BIG-IP® tunnels VXLAN sub-collection resource"""
    def __init__(self, vxlans):
        super(Vxlan, self).__init__(vxlans)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            'tm:net:tunnels:vxlan:vxlanstate'



# coding=utf-8
#
# Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager (GTM) datacenter module.

REST URI
    ``http://localhost/mgmt/tm/gtm/datacenter``

GUI Path
    ``DNS --> GSLB : Data Centers``

REST Kind
    ``tm:gtm:datacenter:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Datacenters(Collection):
    """BIG-IP® GTM datacenter collection"""
    def __init__(self, gtm):
        super(Datacenters, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Datacenter]
        self._meta_data['attribute_registry'] =\
            {'tm:gtm:datacenter:datacenterstate': Datacenter}


class Datacenter(Resource, ExclusiveAttributesMixin):
    """BIG-IP® GTM datacenter resource"""
    def __init__(self, dc_s):
        super(Datacenter, self).__init__(dc_s)
        self._meta_data['required_json_kind'] =\
            'tm:gtm:datacenter:datacenterstate'
        self._meta_data['exclusive_attributes'].append(('enabled', 'disabled'))

    def _endis_attrs(self):
        """Manipulate return value to equal negation of set value

        This function (uniquely?!) manipulates response values before the
        consumer has access to them. We think this is dangerous. It is likely
        this function will move once we figure out how to properly annotate
        this RISKY behavior!"

        The BIG-IP REST API for this particular endpoint has two fields
        which are mutually exclusive; disabled and enabled. When using this
        SDK API, you may do the following

            d = api.tm.gtm.datasources.datasource.load(name='foo')
            d.update(enabled=False)

        You might expect that the behavior of the following...

            if d.enabled:
                print("enabled")
            else:
                print("disabled")

        ...would result in "enabled" being printed, but that would not be
        the case; BIG-IP will specify that "enabled" is True and that the
        following is also now True

            d.disabled == True

        This behavior of setting a different variable instead of the one
        that you specified, may not be obvious to the casual user. Therefore,
        this method will set appropriate sister variables to be the negation
        of the variable you set.

        Therefore

            d.enabled = True

        will also do the following automatically

            d.disabled = False

        This behavior will allow the SDK to behave according to most users
        expectations, shown below

            d.update(enabled=False)
            if d.enabled:
                print("enabled")
            else:
                print("disabled")

        which will print the following

            "disabled"

        Likewise, checking for d.disabled would return True.
        Returns:
            None
        """
        if 'disabled' in self.__dict__:
            self.__dict__['enabled'] = not self.__dict__['disabled']
        if 'enabled' in self.__dict__:
            self.__dict__['disabled'] = not self.__dict__['enabled']
        return None

    def create(self, **kwargs):
        inst = self._create(**kwargs)
        inst._endis_attrs()
        return inst

    def load(self, **kwargs):
        kwargs = self._reduce_boolean_pair(kwargs, 'enabled', 'disabled')
        inst = self._load(**kwargs)
        inst._endis_attrs()
        return inst

    def refresh(self, **kwargs):
        kwargs = self._reduce_boolean_pair(kwargs, 'enabled', 'disabled')
        self._refresh(**kwargs)
        self._endis_attrs()
        return self

    def update(self, **kwargs):
        if 'enabled' in kwargs or 'disabled' in kwargs:
            self.__dict__.pop('enabled')
            self.__dict__.pop('disabled')
        self._update(**kwargs)
        self._endis_attrs()



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® cluster device-group submodule

REST URI
    ``http://localhost/mgmt/tm/cm/device-group``

GUI Path
    ``Device Management --> Device Groups``

REST Kind
    ``tm:cm:device-group:*``
"""

from distutils.version import LooseVersion
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod


def fqdn_name(partition, value):
    if value is not None and not value.startswith('/'):
        return '/{0}/{1}'.format(partition, value)
    return value


class Device_Groups(Collection):
    """BIG-IP® cluster device-groups collection."""
    def __init__(self, cm):
        super(Device_Groups, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [Device_Group]
        self._meta_data['attribute_registry'] =\
            {'tm:cm:device-group:device-groupstate': Device_Group}


class Device_Group(Resource):
    """BIG-IP® cluster device-group resource"""
    def __init__(self, device_groups):
        super(Device_Group, self).__init__(device_groups)
        self._meta_data['read_only_attributes'].append('type')
        self._meta_data['required_json_kind'] =\
            'tm:cm:device-group:device-groupstate'
        self._meta_data['attribute_registry'] = {
            'tm:cm:device-group:devices:devicescollectionstate': Devices_s
        }

    def sync_to(self):
        """Wrapper method that synchronizes configuration to DG.


        Executes the containing object's cm :meth:`~f5.bigip.cm.Cm.exec_cmd`
        method to sync the configuration TO the device-group.

        :note:: Both sync_to, and sync_from methods are convenience
                methods which usually are not what this SDK offers.
                It is best to execute config-sync with the use of
                exec_cmd() method on the cm endpoint.
        """
        device_group_collection = self._meta_data['container']
        cm = device_group_collection._meta_data['container']
        sync_cmd = 'config-sync to-group %s' % self.name
        cm.exec_cmd('run', utilCmdArgs=sync_cmd)

    def sync_from(self):
        """Wrapper method that synchronizes configuration from DG.


        Executes the containing object's cm :meth:`~f5.bigip.cm.Cm.exec_cmd`
        method to sync the configuration FROM the device-group.

        :note:: Both sync_to, and sync_from methods are convenience
                methods which usually are not what this SDK offers.
                It is best to execute config-sync with the use of
                exec_cmd() method on the cm endpoint.
        """
        device_group_collection = self._meta_data['container']
        cm = device_group_collection._meta_data['container']
        sync_cmd = 'config-sync from-group %s' % self.name
        cm.exec_cmd('run', utilCmdArgs=sync_cmd)


class Devices_s(Collection):
    """BIG-IP® cluster devices-group devices subcollection."""
    def __init__(self, device_group):
        super(Devices_s, self).__init__(device_group)
        self._meta_data['allowed_lazy_attributes'] = [Devices]
        self._meta_data['required_json_kind'] =\
            'tm:cm:device-group:devices:devicescollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:cm:device-group:devices:devicesstate': Devices}


class Devices(Resource):
    """BIG-IP® cluster devices-group devices subcollection resource."""
    def __init__(self, devices_s):
        super(Devices, self).__init__(devices_s)
        self._meta_data['required_json_kind'] =\
            'tm:cm:device-group:devices:devicesstate'

    def _fixup_name(self, kwargs):
        # Name munging is required < 11.6.0 and on versions (and sub versions)
        # of 11.6.1.
        # TODO(Remove this when 11.6.1 is no longer supported)
        if 'name' in kwargs:
            kwargs['name'] = fqdn_name('Common', kwargs['name'])
        else:
            self.__dict__['name'] = fqdn_name('Common', self.__dict__['name'])

    def update(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__)

    def modify(self, **kwargs):
        raise UnsupportedMethod(
            "%s does not support the modify method" % self.__class__.__name__)

    def create(self, **kwargs):
        # Name munging is required < 11.6.0 and on versions (and sub versions)
        # of 11.6.1.
        # TODO(Remove this when 11.6.1 is no longer supported)
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        tmos_v = LooseVersion(tmos_v)
        if tmos_v < LooseVersion('11.6.0'):
            self._fixup_name(kwargs)
        elif tmos_v > LooseVersion('11.6.0') and tmos_v < LooseVersion('12.0.0'):
            self._fixup_name(kwargs)
        return self._create(**kwargs)

    def exists(self, **kwargs):
        # Name munging is required < 11.6.0 and on versions (and sub versions)
        # of 11.6.1.
        # TODO(Remove this when 11.6.1 is no longer supported)
        kwargs['partition'] = 'Common'
        return self._exists(**kwargs)

    def delete(self, **kwargs):
        # Name munging is required < 11.6.0 and on versions (and sub versions)
        # of 11.6.1.
        # TODO(Remove this when 11.6.1 is no longer supported)
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        tmos_v = LooseVersion(tmos_v)
        if tmos_v < LooseVersion('11.6.0'):
            self._fixup_name(kwargs)
        elif tmos_v > LooseVersion('11.6.0') and tmos_v < LooseVersion('12.0.0'):
            self._fixup_name(kwargs)
        return self._delete(**kwargs)



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® SNMP submodule.

REST URI
    ``http://localhost/mgmt/tm/sys/snmp/``

GUI Path
    ``System --> SNMP``

REST Kind
    ``tm:snmp:*``
"""
from distutils.version import LooseVersion
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import UnsupportedOperation


class Snmp(UnnamedResource):
    def __init__(self, sys):
        super(Snmp, self).__init__(sys)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] = 'tm:sys:snmp:snmpstate'
        self._meta_data['allowed_lazy_attributes'] = [
            Communities_s,
            Traps_s,
            Users_s
        ]
        self._meta_data['attribute_registry'] = {
            'tm:sys:snmp:communities:communitiescollectionstate':
                Communities_s,
            'tm:sys:snmp:traps:trapscollectionstate': Traps_s,
            'tm:sys:snmp:users:userscollectionstate': Users_s
        }


class Communities_s(Collection):
    """BIG-IP® SNMP Communities collection."""
    def __init__(self, snmp):
        super(Communities_s, self).__init__(snmp)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:communities:communitiescollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Community]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:snmp:communities:communitiesstate': Community}


class Community(Resource):
    """BIG-IP® SNMP Community resource."""
    def __init__(self, communities_s):
        super(Community, self).__init__(communities_s)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:communities:communitiesstate'
        self._meta_data['required_creation_parameters']\
            .update(('communityName',))


class Traps_s(Collection):
    """BIG-IP® SNMP Traps collection."""
    def __init__(self, snmp):
        super(Traps_s, self).__init__(snmp)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:traps:trapscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Trap]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:snmp:traps:trapsstate': Trap}


class Trap(Resource):
    """BIG-IP® SNMP Trap resource."""
    def __init__(self, traps_s):
        super(Trap, self).__init__(traps_s)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:traps:trapsstate'

    def update(self, **kwargs):
        """Due to a password decryption bug

        we will disable update() method for 12.1.0 and up

        """
        tmos_version = self._meta_data['bigip'].tmos_version
        if LooseVersion(tmos_version) > LooseVersion('12.0.0'):
            msg = "Update() is unsupported for Trap on version %s. " \
                  "Utilize Modify() method instead" % tmos_version
            raise UnsupportedOperation(msg)
        else:
            self._update(**kwargs)


class Users_s(Collection):
    """BIG-IP® SNMP Users collection."""
    def __init__(self, snmp):
        super(Users_s, self).__init__(snmp)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:users:userscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [User]
        self._meta_data['attribute_registry'] = \
            {'tm:sys:snmp:users:usersstate': User}


class User(Resource):
    """BIG-IP® SNMP User resource."""
    def __init__(self, users_s):
        super(User, self).__init__(users_s)
        self._meta_data['required_json_kind'] = \
            'tm:sys:snmp:users:usersstate'
        self._meta_data['required_creation_parameters'] \
            .update(('authProtocol',))

    def update(self, **kwargs):
        """Due to a password decryption bug

        we will disable update() method for 12.1.0 and up

        """
        tmos_version = self._meta_data['bigip'].tmos_version
        if LooseVersion(tmos_version) > LooseVersion('12.0.0'):
            msg = "Update() is unsupported for User on version %s. " \
                  "Utilize Modify() method instead" % tmos_version
            raise UnsupportedOperation(msg)
        else:
            self._update(**kwargs)



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® user module

REST URI
    ``http://localhost/mgmt/auth/user/``

GUI Path
    ``System --> Users``

REST Kind
    ``tm:auth:user:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Users(Collection):
    """BIG-IP® user collection"""
    def __init__(self, auth):
        super(Users, self).__init__(auth)
        self._meta_data['allowed_lazy_attributes'] = [User]
        self._meta_data['attribute_registry'] = \
            {'tm:auth:user:userstate': User}


class User(Resource):
    """BIG-IP® user resource"""
    def __init__(self, users):
        super(User, self).__init__(users)
        self._meta_data['required_json_kind'] = 'tm:auth:user:userstate'



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Network ARP module.

REST URI
    ``http://localhost/mgmt/tm/net/dns-resolver``

GUI Path
    ``Network --> Dns Resolvers``

REST Kind
    ``tm:net:dns:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Dns_Resolvers(Collection):
    """BIG-IP® network Dns Resolver collection"""
    def __init__(self, net):
        super(Dns_Resolvers, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Dns_Resolver]
        self._meta_data['attribute_registry'] = {
            'tm:net:dns-resolver:dns-resolverstate': Dns_Resolver
        }


class Dns_Resolver(Resource):
    """BIG-IP® Dns Resolver resource."""
    def __init__(self, Dns_Resolvers):
        super(Dns_Resolver, self).__init__(Dns_Resolvers)
        self._meta_data['required_json_kind'] = \
            'tm:net:dns-resolver:dns-resolverstate'



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# NOTE:  Code taken from Effective Python Item 26


try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError as exc:
        message = ("Maybe you're using Python < 2.7 and do not have the "
                   "orderreddict external dependency installed.")
        raise exc(message)

from distutils.version import LooseVersion
from six import iteritems

import logging

from f5.sdk_exception import EmptyContent
from f5.sdk_exception import InvalidCommand
from f5.sdk_exception import LazyAttributesRequired
from f5.sdk_exception import MissingHttpHeader
from f5.sdk_exception import UnsupportedMethod
from f5.sdk_exception import UnsupportedTmosVersion
from f5.sdk_exception import UtilError


class ToDictMixin(object):
    """Convert an object's attributes to a dictionary"""
    traversed = {}
    Containers = tuple, list, set, frozenset, dict

    def to_dict(self):
        ToDictMixin.traversed = {}
        return self._to_dict()

    def _to_dict(self):
        result = self._traverse_dict(self.__dict__)
        return result

    def _traverse_dict(self, instance_dict):
        output = {}

        # This iteration breaks if the second value comes before
        # the first. We must use ordered dicts here
        tmp = OrderedDict(sorted(iteritems(instance_dict), key=lambda t: t[0]))
        for key, value in iteritems(tmp):
            output[key] = self._traverse(key, value)
        return output

    def _traverse(self, key, value):
        if isinstance(value, ToDictMixin.Containers) or\
           hasattr(value, '__dict__'):
            if id(value) in ToDictMixin.traversed:
                return ToDictMixin.traversed[id(value)]
            else:
                ToDictMixin.traversed[id(value)] = ['TraversalRecord', key]

        if isinstance(value, ToDictMixin):
            return value._to_dict()
        elif isinstance(value, dict):
            return self._traverse_dict(value)
        elif isinstance(value, list):
            return [self._traverse(key, item) for item in value]
        elif hasattr(value, '__dict__'):
            return self._traverse_dict(value.__dict__)
        else:
            return value


class LazyAttributeMixin(object):
    """Allow attributes to be created lazily based on the allowed values"""
    def __getattr__(container, name):
        # ensure this object supports lazy attrs.
        cls_name = container.__class__.__name__
        if 'allowed_lazy_attributes' not in container._meta_data:
            error_message = ('"allowed_lazy_attributes" not in',
                             'container._meta_data for class %s' % cls_name)
            raise LazyAttributesRequired(error_message)

        # ensure the requested attr is present
        attr_names = container.transform_attr_names()
        if name not in attr_names:
            error_message = "'%s' object has no attribute '%s'"\
                % (container.__class__, name)
            raise AttributeError(error_message)

        # Instantiate and potentially set the attr on the object
        # Issue #112 -- Only call setattr here if the lazy attribute
        # is NOT a `Resource`.  This should allow for only 1 ltm attribute
        # but many nat attributes just like the BIGIP device.
        for lazy_attribute in container._meta_data['allowed_lazy_attributes']:
            if name == lazy_attribute.__name__.lower():
                attribute = lazy_attribute(container)
                bases = [base.__name__ for base in lazy_attribute.__bases__]
                # Doing version check per each resource
                container._check_supported_versions(container, attribute)
                if 'Resource' not in bases:
                    setattr(container, name, attribute)
                return attribute

    def transform_attr_names(self):
        attr_names = \
            [la.__name__.lower() for la in
             self._meta_data['allowed_lazy_attributes']]
        return attr_names

    def _check_supported_versions(self, container, attribute):
        tmos_v = container._meta_data['bigip'].tmos_version
        minimum = attribute._meta_data['minimum_version']
        if LooseVersion(tmos_v) < LooseVersion(minimum):
            error = "There was an attempt to access resource: \n{}\n which " \
                    "is not implemented in the device's TMOS version: {}. " \
                    "The minimum TMOS version in which this resource *is* " \
                    "supported is {}".format(
                        attribute._meta_data['uri'],
                        tmos_v,
                        minimum)
            raise UnsupportedTmosVersion(error)

    def _is_version_supported_method(container, method_version):
        """Helper method

         To use in instances where class methods on some resources
         require a specific TMOS version to run.

        Raises::
                UnsupportedTmosVersion
        """
        tmos_v = container._meta_data['bigip'].tmos_version
        if LooseVersion(tmos_v) < LooseVersion(method_version):
            error = "There was an attempt to use a method which " \
                    "has not been implemented or supported " \
                    "in the device's TMOS version: %s. " \
                    "Minimum TMOS version supported is %s" % (
                        tmos_v, method_version)
            raise UnsupportedTmosVersion(error)


class ExclusiveAttributesMixin(object):
    """Overrides ``__setattr__`` to remove exclusive attrs from the object."""
    def __setattr__(self, key, value):
        """Remove any of the existing exclusive attrs from the object

        Objects attributes can be exclusive for example disable/enable.  So
        we need to make sure objects only have one of these attributes at
        at time so that the updates won't fail.
        """
        if '_meta_data' in self.__dict__:
            # Sometimes this is called prior to full object construction
            for attr_set in self._meta_data['exclusive_attributes']:
                if key in attr_set:
                    new_set = set(attr_set) - set([key])
                    [self.__dict__.pop(n, '') for n in new_set]
        # Now set the attribute
        super(ExclusiveAttributesMixin, self).__setattr__(key, value)


class CommandExecutionMixin(object):
    """This adds command execution option on the objects.

    These objects do not support create, delete, load, and require
    a separate method of execution. Commands do not have
    direct mapping to an HTTP method so usage of POST
    and an absolute URI is required.

    """

    def create(self, **kwargs):
        """Create is not supported for command execution

        :raises: UnsupportedOperation
        """
        raise UnsupportedMethod(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for command execution

        :raises: UnsupportedOperation
        """
        raise UnsupportedMethod(
            "%s does not support the delete method" % self.__class__.__name__
        )

    def load(self, **kwargs):
        """Load is not supported for command execution

                :raises: UnsupportedOperation
        """
        raise UnsupportedMethod(
            "%s does not support the load method" % self.__class__.__name__
        )

    def _is_allowed_command(self, command):
        """Checking if the given command is allowed on a given endpoint."""
        cmds = self._meta_data['allowed_commands']
        if command not in self._meta_data['allowed_commands']:
            error_message = "The command value {0} does not exist. " \
                            "Valid commands are {1}".format(command, cmds)
            raise InvalidCommand(error_message)

    def _check_command_result(self):
        """If command result exists run these checks."""
        if self.commandResult.startswith('/bin/bash'):
            raise UtilError('%s' % self.commandResult.split(' ', 1)[1])
        if self.commandResult.startswith('/bin/mv'):
            raise UtilError('%s' % self.commandResult.split(' ', 1)[1])
        if self.commandResult.startswith('/bin/ls'):
            raise UtilError('%s' % self.commandResult.split(' ', 1)[1])
        if self.commandResult.startswith('/bin/rm'):
            raise UtilError('%s' % self.commandResult.split(' ', 1)[1])
        if 'invalid option' in self.commandResult:
            raise UtilError('%s' % self.commandResult)
        if 'Invalid option' in self.commandResult:
            raise UtilError('%s' % self.commandResult)
        if 'usage: /usr/bin/get_dossier' in self.commandResult:
            raise UtilError('%s' % self.commandResult)

    def exec_cmd(self, command, **kwargs):
        """Wrapper method that can be changed in the inheriting classes."""
        self._is_allowed_command(command)
        self._check_command_parameters(**kwargs)
        return self._exec_cmd(command, **kwargs)

    def _exec_cmd(self, command, **kwargs):
        """Create a new method as command has specific requirements.

        There is a handful of the TMSH global commands supported,
        so this method requires them as a parameter.

        :raises: InvalidCommand
        """

        kwargs['command'] = command
        self._check_exclusive_parameters(**kwargs)
        requests_params = self._handle_requests_params(kwargs)
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.post(
            self._meta_data['uri'], json=kwargs, **requests_params)
        new_instance = self._stamp_out_core()
        new_instance._local_update(response.json())
        if 'commandResult' in new_instance.__dict__:
            new_instance._check_command_result()

        return new_instance


class FileUploadMixin(object):
    def _upload_file(self, filepathname, **kwargs):
        with open(filepathname, 'rb') as fileobj:
            self._upload(fileobj, **kwargs)

    def _upload(self, fileinterface, **kwargs):
        size = len(fileinterface.read())
        fileinterface.seek(0)
        requests_params = self._handle_requests_params(kwargs)
        session = self._meta_data['icr_session']
        chunk_size = kwargs.pop('chunk_size', 512 * 1024)
        start = 0
        while True:
            file_slice = fileinterface.read(chunk_size)
            if not file_slice:
                break

            current_bytes = len(file_slice)
            if current_bytes < chunk_size:
                end = size
            else:
                end = start + current_bytes
            headers = {
                'Content-Range': '%s-%s/%s' % (start, end - 1, size),
                'Content-Type': 'application/octet-stream'}
            data = {
                'data': file_slice,
                'headers': headers,
                'verify': False
            }
            logging.debug(data)
            requests_params.update(data)
            session.post(self.file_bound_uri,
                         **requests_params)
            start += current_bytes


class FileDownloadMixin(object):
    def _download_file(self, src, dest, **kwargs):
        with open(dest, 'wb') as fileobj:
            self._download(src, fileobj, **kwargs)

    def _download(self, src, fileinterface, **kwargs):
        requests_params = self._handle_requests_params(kwargs)
        session = self._meta_data['icr_session']
        chunk_size = kwargs.pop('chunk_size', 512 * 1024)
        self.file_bound_uri = self._meta_data['uri'] + src
        start = 0
        end = chunk_size - 1
        size = 0
        current_bytes = 0

        while True:
            content_range = "%s-%s/%s" % (start, end, size)
            headers = {
                'Content-Range': content_range,
                'Content-Type': 'application/octet-stream'
            }
            data = {
                'headers': headers,
                'verify': False,
                'stream': True
            }
            logging.debug(data)
            requests_params.update(data)
            response = session.get(self.file_bound_uri,
                                   **requests_params)
            if response.status_code == 200:
                # If the size is zero, then this is the first time through
                # the loop and we don't want to write data because we
                # haven't yet figured out the total size of the file.
                if size > 0:
                    current_bytes += chunk_size
                    for chunk in response.iter_content(chunk_size):
                        fileinterface.write(chunk)
            # Once we've downloaded the entire file, we can break out of
            # the loop
            if end == size:
                break
            crange = response.headers['Content-Range']
            # Determine the total number of bytes to read.
            if size == 0:
                size = int(crange.split('/')[-1]) - 1
                # If the file is smaller than the chunk_size, the BigIP
                # will return an HTTP 400. Adjust the chunk_size down to
                # the total file size...
                if chunk_size > size:
                    end = size
                # ...and pass on the rest of the code.
                continue
            start += chunk_size
            if (current_bytes + chunk_size) > size:
                end = size
            else:
                end = start + chunk_size - 1


class AsmFileMixin(object):
    """Mixin for manipulating files for ASM file-transfer endpoints.


    For ease of code maintenance this is separate from FileUploadMixin
    on purpose.

    """
    def _download_file(self, filepathname):
        self._download(filepathname)

    def _download(self, filepathname):
        session = self._meta_data['icr_session']
        with open(filepathname, 'wb') as writefh:
            headers = {
                'Content-Type': 'application/json'
            }
            req_params = {'headers': headers,
                          'verify': False}
            response = session.get(self.file_bound_uri, **req_params)
            if response.status_code == 200:
                if 'Content-Length' not in response.headers:
                    error_message = "The Content-Length header is not present."
                    raise MissingHttpHeader(error_message)

                length = response.headers['Content-Length']

                if int(length) > 0:
                    writefh.write(response.content)
                else:
                    error = "Invalid Content-Length value returned: %s ," \
                            "the value should be greater than 0" % length
                    raise EmptyContent(error)

    def _upload_file(self, filepathname, **kwargs):
        with open(filepathname, 'rb') as fileobj:
            self._upload(fileobj, **kwargs)

    def _upload(self, fileinterface, **kwargs):
        size = len(fileinterface.read())
        fileinterface.seek(0)
        requests_params = self._handle_requests_params(kwargs)
        session = self._meta_data['icr_session']
        chunk_size = kwargs.pop('chunk_size', 512 * 1024)
        start = 0
        while True:
            file_slice = fileinterface.read(chunk_size)
            if not file_slice:
                break

            current_bytes = len(file_slice)
            if current_bytes < chunk_size:
                end = size
            else:
                end = start + current_bytes
            headers = {
                'Content-Range': '%s-%s/%s' % (start, end - 1, size),
                'Content-Type': 'application/octet-stream'}
            data = {'data': file_slice,
                    'headers': headers,
                    'verify': False}
            logging.debug(data)
            requests_params.update(data)
            session.post(self.file_bound_uri,
                         **requests_params)
            start += current_bytes


class DeviceMixin(object):
    '''Manage BigIP device cluster in a general way.'''

    def get_device_info(self, bigip):
        '''Get device information about a specific BigIP device.

        :param bigip: bigip object --- device to inspect
        :returns: bigip object
        '''

        coll = bigip.tm.cm.devices.get_collection()
        device = [device for device in coll if device.selfDevice == 'true']
        assert len(device) == 1
        return device[0]


class CheckExistenceMixin(object):
    '''In 11.6.0 some items return True on exists whether they exist or not'''

    def _check_existence_by_collection(self, container, item_name):
        '''Check existnce of item based on get collection call.

        :param collection: container object -- capable of get_collection()
        :param item_name: str -- name of item to search for in collection
        '''

        coll = container.get_collection()
        for item in coll:
            if item.name == item_name:
                return True
        return False

    def _return_object(self, container, item_name):
        """Helper method to retrieve the object"""
        coll = container.get_collection()
        for item in coll:
            if item.name == item_name:
                return item


class UpdateMonitorMixin(object):
    def update(self, **kwargs):
        """Change the configuration of the resource on the device.

        This method uses Http PUT alter the service state on the device.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.  Various edge cases are handled:

        * read-only attributes that are unchangeable are removed
        * ``defaultsFrom`` attribute is removed from JSON before the PUT

        :param kwargs: keys and associated values to alter on the device

        """
        self.__dict__.pop('defaultsFrom', '')
        self._update(**kwargs)



# coding=utf-8
#
# Copyright 2015-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®)  global-settings submodule.

REST URI
    ``http://localhost/mgmt/tm/gtm/global-settings``

GUI Path
    ``DNS --> Settings --> GSLB``

REST Kind
    ``tm:gtm:global-settings*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import UnnamedResource


class Global_Settings(OrganizingCollection):
    """BIG-IP® GTM global-settings organizing collection."""
    def __init__(self, gtm):
        super(Global_Settings, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [
            General,
            Load_Balancing,
            Metrics,
        ]


class General(UnnamedResource):
    """BIG-IP® GTM global-settings general resource."""
    def __init__(self, global_settings):
        super(General, self).__init__(global_settings)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:global-settings:general:generalstate'
        self._meta_data['required_load_parameters'] = set()


class Load_Balancing(UnnamedResource):
    """BIG-IP® GTM global-settings load balancing resource."""
    def __init__(self, global_settings):
        super(Load_Balancing, self).__init__(global_settings)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:global-settings:load-balancing:load-balancingstate'
        self._meta_data['required_load_parameters'] = set()


class Metrics(UnnamedResource):
    """BIG-IP® GTM global-settings metrics resource."""
    def __init__(self, global_settings):
        super(Metrics, self).__init__(global_settings)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:global-settings:metrics:metricsstate'
        self._meta_data['required_load_parameters'] = set()



# coding=utf-8
#
#  Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®) pool module.

REST URI
    ``http://localhost/mgmt/tm/gtm/pool``

GUI Path
    ``DNS --> GSLB : Pools``

REST Kind
    ``tm:gtm:pools:*``
"""

from distutils.version import LooseVersion

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import URICreationCollision
from requests.exceptions import HTTPError


class Pools(object):
    """BIG-IP® GTM Pool factory

    The GTM Pool factory object is used to provide a consistent user
    experience across the SDK, while supporting the breaking changes in
    the BIG-IP APIs.

    Between the 11.x and 12.x releases of BIG-IP, the REST endpoint for
    Pool changed from a Collection to an OrganizingCollection. The
    result is that breaking changes occurred in the API.

    This breakage led to a discussion on naming conventions because there
    appeared to be a conflict now. For example, depending on your version,
    only one of the following would work.

    For 11.x,

        tm.gtm.pools.pool.load(name='foo')

    For 12.x,

        tm.gtm.pools.a_s.a.load(name='foo')

    but not both. To stick with a consistent user experience, we decided
    to override the __new__ method to support both examples above. The SDK
    automatically detects which one to use based on the BIG-IP you are
    communicating with.
    """
    def __new__(cls, container):
        tmos_v = container._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) < LooseVersion('12.0.0'):
            return PoolCollection(container)
        else:
            return PoolOrganizingCollection(container)


class Members_s(object):
    """BIG-IP® GTM Members_s factory

    The GTM Members_s factory is used here for code readability
    and maintenance purposes, to allow us to stay in conventions already
    set in this SDK, and at the same time accommodate changes between v11
    and v12 versions of Members_s sub-collection
    """
    def __new__(cls, container):
        tmos_v = container._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) < LooseVersion('12.0.0'):
            return MembersCollection_v11(container)
        else:
            if isinstance(container, A):
                return MembersCollectionA(container)
            if isinstance(container, Aaaa):
                return MembersCollectionAAAA(container)
            if isinstance(container, Cname):
                return MembersCollectionCname(container)
            if isinstance(container, Mx):
                return MembersCollectionMx(container)
            if isinstance(container, Naptr):
                return MembersCollectionNaptr(container)
            if isinstance(container, Srv):
                return MembersCollectionSrv(container)


class MembersCollection_v11(Collection):
    """v11.x BIG-IP® GTM pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollection_v11, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:members:membersstate': Member}


class MembersCollectionA(Collection):
    """v12.x BIG-IP® GTM A pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionA, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:a:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:a:members:membersstate': Member}


class MembersCollectionAAAA(Collection):
    """v12.x BIG-IP® GTM AAAA pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionAAAA, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:aaaa:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:aaaa:members:membersstate': Member}


class MembersCollectionCname(Collection):
    """v12.x BIG-IP® GTM CNAME pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionCname, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:cname:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:cname:members:membersstate': Member}


class MembersCollectionMx(Collection):
    """v12.x BIG-IP® GTM MX pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionMx, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:mx:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:mx:members:membersstate': Member}


class MembersCollectionNaptr(Collection):
    """v12.x BIG-IP® GTM NAPTR pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionNaptr, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:naptr:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:naptr:members:membersstate': Member}


class MembersCollectionSrv(Collection):
    """v12.x BIG-IP® GTM SRV pool members sub-collection"""
    def __init__(self, pool):
        self.__class__.__name__ = 'Members_s'
        super(MembersCollectionSrv, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Member]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:srv:members:memberscollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:srv:members:membersstate': Member}


class Member(object):
    """BIG-IP® GTM Members factory

    The GTM Members factory is used here for code readability
    and maintenance purposes, to allow us to stay in conventions already
    set in this SDK, and at the same time accommodate changes between v11
    and v12 versions of Members resource
    """
    def __new__(cls, container):
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:members:memberscollectionstate':
            return MembersResource_v11(container)
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:a:members:memberscollectionstate':
            return MembersResourceA(container)
        if container._meta_data['required_json_kind'] ==  \
                'tm:gtm:pool:aaaa:members:memberscollectionstate':
            return MembersResourceAAAA(container)
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:cname:members:memberscollectionstate':
            return MembersResourceCname(container)
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:mx:members:memberscollectionstate':
            return MembersResourceMx(container)
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:naptr:members:memberscollectionstate':
            return MembersResourceNaptr(container)
        if container._meta_data['required_json_kind'] == \
                'tm:gtm:pool:srv:members:memberscollectionstate':
            return MembersResourceSrv(container)


class MembersResource_v11(Resource):
    """v11.x BIG-IP® GTM members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResource_v11, self).__init__(members_s)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:members:membersstate'


class MembersResourceA(Resource):
    """v12.x BIG-IP® GTM A pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceA, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:a:members:membersstate'
        self._meta_data['required_creation_parameters'].update(('partition',))

    def create(self, **kwargs):
        """This method needs to be created due to a bug in 12.x

        The issue arises when you attempt to create a members sub-collection
        resource and while the operation succeeds BIGIP responds with 404.

        By requiring partition as argument we have shielded ourselves from
        this issue in 12.0.0, however in v12.1.x the problem occurs, even when
        the partition parameter is submitted. Custom create method is required
        to catch and deal with iControlUnexpectedHTTPError exception.

        Issue is tracked under: ID610441.

        note:: If version is 12.x then we use this method, otherwise we use
               _create method from Resource class as usual.
        """
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) >= LooseVersion('12.1.0'):
            if 'uri' in self._meta_data:
                error = "There was an attempt to assign a new uri to this " \
                        "resource, the _meta_data['uri'] is %s and it should" \
                        " not be changed." % (self._meta_data['uri'])
                raise URICreationCollision(error)
            self._check_exclusive_parameters(**kwargs)
            requests_params = self._handle_requests_params(kwargs)
            self._check_create_parameters(**kwargs)

            # Reduce boolean pairs as specified by the meta_data entry below
            for key1, key2 in self._meta_data['reduction_forcing_pairs']:
                kwargs = self._reduce_boolean_pair(kwargs, key1, key2)

            # Make convenience variable with short names for this method.
            _create_uri = self._meta_data['container']._meta_data['uri']
            session = self._meta_data['bigip']._meta_data['icr_session']
            # This is a bit hacky but we need to do this so we are able to
            # create a resource inside SDK properly. We also include the
            # scenario where 20x range response occurs just in case this gets
            # fixed in later 12.x release.

            try:
                response = session.post(
                    _create_uri, json=kwargs, **requests_params)

            except HTTPError as err:
                if err.response.status_code != 404:
                    raise
                if err.response.status_code == 404:
                    kwargs['uri_as_parts'] = True
                    response = session.get(_create_uri, **kwargs)
                    return self._produce_instance(response)

            # Make new instance of self
            return self._produce_instance(response)

        else:
            return self._create(**kwargs)


class MembersResourceAAAA(Resource):
    """v12.x BIG-IP® GTM AAAA pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceAAAA, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:aaaa:members:membersstate'
        self._meta_data['required_creation_parameters'].update(('partition',))

    def create(self, **kwargs):
        """This method needs to be created due to a bug in 12.x

        The issue arises when you attempt to create a members sub-collection
        resource and while the operation succeeds BIGIP responds with 404.

        By requiring partition as argument we have shielded ourselves from
        this issue in 12.0.0, however in v12.1.x the problem occurs, even when
        the partition parameter is submitted. Custom create method is required
        to catch and deal with iControlUnexpectedHTTPError exception.

        Issue is tracked under: ID610441.

        note:: If version is 12.x then we use this method, otherwise we use
               _create method from Resource class as usual.
        """
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) >= LooseVersion('12.1.0'):
            if 'uri' in self._meta_data:
                error = "There was an attempt to assign a new uri to this " \
                        "resource, the _meta_data['uri'] is %s and it should" \
                        " not be changed." % (self._meta_data['uri'])
                raise URICreationCollision(error)
            self._check_exclusive_parameters(**kwargs)
            requests_params = self._handle_requests_params(kwargs)
            self._check_create_parameters(**kwargs)

            # Reduce boolean pairs as specified by the meta_data entry below
            for key1, key2 in self._meta_data['reduction_forcing_pairs']:
                kwargs = self._reduce_boolean_pair(kwargs, key1, key2)

            # Make convenience variable with short names for this method.
            _create_uri = self._meta_data['container']._meta_data['uri']
            session = self._meta_data['bigip']._meta_data['icr_session']
            # This is a bit hacky but we need to do this so we are able to
            # create a resource inside SDK properly. We also include the
            # scenario where 20x range response occurs just in case this gets
            # fixed in later 12.x release.

            try:
                response = session.post(
                    _create_uri, json=kwargs, **requests_params)

            except HTTPError as err:
                if err.response.status_code != 404:
                    raise
                if err.response.status_code == 404:
                    kwargs['uri_as_parts'] = True
                    response = session.get(_create_uri, **kwargs)
                    return self._produce_instance(response)

            # Make new instance of self
            return self._produce_instance(response)

        else:
            return self._create(**kwargs)


class MembersResourceCname(Resource):
    """v12.x BIG-IP® GTM CNAME pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceCname, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:cname:members:membersstate'


class MembersResourceMx(Resource):
    """v12.x BIG-IP® GTM MX pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceMx, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:mx:members:membersstate'


class MembersResourceNaptr(Resource):
    """v12.x BIG-IP® GTM NAPTR pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceNaptr, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:naptr:members:membersstate'
        self._meta_data['required_creation_parameters'].update(
            ('flags', 'service'))


class MembersResourceSrv(Resource):
    """v12.x BIG-IP® GTM SRV pool members resource"""
    def __init__(self, members_s):
        self.__class__.__name__ = 'Member'
        super(MembersResourceSrv, self).__init__(members_s)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:srv:members:membersstate'
        self._meta_data['required_creation_parameters'].update(
            ('port',))


class PoolCollection(Collection):
    """v11.x BIG-IP® GTM pool collection"""
    def __init__(self, gtm):
        self.__class__.__name__ = 'Pools'
        super(PoolCollection, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Pool]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:poolstate': Pool}


class Pool(Resource):
    """v11.x BIG-IP® GTM pool resource"""
    def __init__(self, pool_s):
        super(Pool, self).__init__(pool_s)
        self._meta_data['required_json_kind'] = 'tm:gtm:pool:poolstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:memberscollectionstate': Members_s
        }


class PoolOrganizingCollection(OrganizingCollection):
    """v12.x GTM pool is an OC."""
    def __init__(self, gtm):
        self.__class__.__name__ = 'Pool'
        super(PoolOrganizingCollection, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [
            A_s,
            Aaaas,
            Cnames,
            Mxs,
            Naptrs,
            Srvs,
            ]


class A_s(Collection):
    """v12.x BIG-IP® A pool collection.

    Class name needed changing due to
    'as' being reserved python keyword
    """
    def __init__(self, pool):
        super(A_s, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [A]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:a:astate': A}


class A(Resource):
    """v12.x BIG-IP®A pool resource"""
    def __init__(self, _as):
        super(A, self).__init__(_as)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:a:astate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:a:members:memberscollectionstate': Members_s
        }


class Aaaas(Collection):
    """v12.x BIG-IP® AAAA pool collection"""
    def __init__(self, pool):
        super(Aaaas, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Aaaa]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:aaaa:aaaastate': Aaaa}


class Aaaa(Resource):
    """v12.x BIG-IP® AAAA pool resource"""
    def __init__(self, aaaas):
        super(Aaaa, self).__init__(aaaas)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:aaaa:aaaastate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:aaaa:members:memberscollectionstate': Members_s
        }


class Cnames(Collection):
    """v12.x BIG-IP® CNAME pool collection"""
    def __init__(self, pool):
        super(Cnames, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Cname]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:cname:cnamestate': Cname}


class Cname(Resource):
    """v12.x BIG-IP® CNAME pool resource"""
    def __init__(self, cnames):
        super(Cname, self).__init__(cnames)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:cname:cnamestate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:cname:members:memberscollectionstate': Members_s
        }


class Mxs(Collection):
    """v12.x BIG-IP® MX pool collection."""
    def __init__(self, pool):
        super(Mxs, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Mx]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:mx:mxstate': Mx}


class Mx(Resource):
    """v12.x BIG-IP® MX pool resource"""
    def __init__(self, mxs):
        super(Mx, self).__init__(mxs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:mx:mxstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:mx:members:memberscollectionstate': Members_s
        }


class Naptrs(Collection):
    """v12.x BIG-IP® NAPTR pool collection"""
    def __init__(self, pool):
        super(Naptrs, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Naptr]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:naptr:naptrstate': Naptr}


class Naptr(Resource):
    """v12.x BIG-IP® NAPTR pool resource"""
    def __init__(self, naptrs):
        super(Naptr, self).__init__(naptrs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:naptr:naptrstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:naptr:members:memberscollectionstate': Members_s
        }


class Srvs(Collection):
    """v12.x BIG-IP® SRV pool collection"""
    def __init__(self, pool):
        super(Srvs, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Srv]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:pool:srv:srvstate': Srv}


class Srv(Resource):
    """v12.x BIG-IP® SRV pool resource"""
    def __init__(self, naptrs):
        super(Srv, self).__init__(naptrs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:pool:srv:srvstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:pool:srv:members:memberscollectionstate': Members_s
        }



# coding=utf-8
#
"""Classes and functions for configuring BIG-IP"""
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip.cm.autodeploy.software_images import Software_Image_Downloads
from f5.bigip.cm.autodeploy.software_images import Software_Image_Uploads
from f5.bigip.resource import OrganizingCollection


class Autodeploy(OrganizingCollection):
    """An organizing collection for Autodeploy resources."""
    def __init__(self, cm):
        super(Autodeploy, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [
            Software_Image_Downloads,
            Software_Image_Uploads
        ]



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection


class Json_Profiles_s(Collection):
    """BIG-IP® ASM Json-Profiles sub-collection.

    Due to the bug that prevents from creating this object in 11.5.4 Final,
    I am disabling this for anything lower than 11.6.0.
    This will be subject to change at some point
    """
    def __init__(self, policy):
        super(Json_Profiles_s, self).__init__(policy)
        self._meta_data['minimum_version'] = '11.6.0'
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Json_Profile]
        self._meta_data['required_json_kind'] = 'tm:asm:policies:json-profiles:json-profilecollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:json-profiles:json-profilestate': Json_Profile
        }


class Json_Profile(AsmResource):
    """BIG-IP® ASM Json-Profile resource."""
    def __init__(self, json_profiles_s):
        super(Json_Profile, self).__init__(json_profiles_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:json-profiles:json-profilestate'



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager (LTM) persistence module.

REST URI
    ``https://localhost/mgmt/tm/ltm/persistence/``

GUI Path
    ``Local Traffic --> Profiles --> Persistence``

REST Kind
    ``tm:ltm:persistence:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Persistence(OrganizingCollection):
    """BIG-IP® LTM persistence collection."""
    def __init__(self, ltm):
        super(Persistence, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [
            Cookies,
            Dest_Addrs,
            Hashs,
            Msrdps,
            Sips,
            Source_Addrs,
            Ssls,
            Universals
        ]


class Source_Addrs(Collection):
    """BIG-IP® Source Address persistence collection."""
    def __init__(self, persistence):
        super(Source_Addrs, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Source_Addr]
        self._meta_data['attribute_registry'] = \
            {'tm:ltm:persistence:source-addr:source-addrstate': Source_Addr}


class Source_Addr(Resource):
    """BIG-IP® Source Address persistence resource."""
    def __init__(self, source_addrs):
        super(Source_Addr, self).__init__(source_addrs)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:persistence:source-addr:source-addrstate'


class Hashs(Collection):
    """BIG-IP® Hash persistence collection."""
    def __init__(self, persistence):
        super(Hashs, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Hash]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:persistence:hash:hashstate': Hash}


class Hash(Resource):
    """BIG-IP® Hash persistence resource."""
    def __init__(self, hashs):
        super(Hash, self).__init__(hashs)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:persistence:hash:hashstate'


class Sips(Collection):
    """BIG-IP® Sip persistence collection."""
    def __init__(self, persistence):
        super(Sips, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Sip]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:persistence:sip:sipstate': Sip}


class Sip(Resource):
    """BIG-IP® Sip persistence resource."""
    def __init__(self, sips):
        super(Sip, self).__init__(sips)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:persistence:sip:sipstate'


class Ssls(Collection):
    """BIG-IP® SSL persistence collection."""
    def __init__(self, persistence):
        super(Ssls, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Ssl]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:persistence:ssl:sslstate': Ssl}


class Ssl(Resource):
    """BIG-IP® SSL persistence resource."""
    def __init__(self, ssls):
        super(Ssl, self).__init__(ssls)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:persistence:ssl:sslstate'


class Global_Settings(Resource):
    """BIG-IP® Global-Settings persistence resource."""
    def __init__(self, Global_Settings_s):
        super(Global_Settings, self).__init__(Global_Settings_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:persistence:global-settings:global-settingsstate'


class Dest_Addrs(Collection):
    """BIG-IP® Destination Address persistence collection."""
    def __init__(self, persistence):
        super(Dest_Addrs, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Dest_Addr]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:persistence:dest-addr:dest-addrstate': Dest_Addr}


class Dest_Addr(Resource):
    """BIG-IP® Destination Address persistence resource."""
    def __init__(self, dest_addrs):
        super(Dest_Addr, self).__init__(dest_addrs)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:persistence:dest-addr:dest-addrstate'


class Msrdps(Collection):
    """BIG-IP® MS RDP persistence collection."""
    def __init__(self, persistence):
        super(Msrdps, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Msrdp]
        self._meta_data['attribute_registry'] = {
            'tm:ltm:persistence:msrdp:msrdpstate': Msrdp}


class Msrdp(Resource):
    """BIG-IP® MS RDP persistence resource."""
    def __init__(self, msrdps):
        super(Msrdp, self).__init__(msrdps)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:persistence:msrdp:msrdpstate'


class Cookies(Collection):
    """BIG-IP® Cookie persistence collection."""
    def __init__(self, persistence):
        super(Cookies, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Cookie]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:persistence:cookie:cookiestate': Cookie}


class Cookie(Resource):
    """BIG-IP® Cookie persistence resource."""
    def __init__(self, cookies):
        super(Cookie, self).__init__(cookies)
        self._meta_data['required_json_kind'] =\
            "tm:ltm:persistence:cookie:cookiestate"


class Universals(Collection):
    """BIG-IP® Universal persistence collection."""
    def __init__(self, persistence):
        super(Universals, self).__init__(persistence)
        self._meta_data['allowed_lazy_attributes'] = [Universal]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:persistence:universal:universalstate': Universal}


class Universal(Resource):
    """BIG-IP® Universal persistence resource."""
    def __init__(self, universals):
        super(Universal, self).__init__(universals)
        self._meta_data['required_json_kind'] =\
            "tm:ltm:persistence:universal:universalstate"



# coding=utf-8
#
#  Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedMethod


class System(OrganizingCollection):
    def __init__(self, shared):
        super(System, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Authn
        ]


class Authn(OrganizingCollection):
    def __init__(self, system):
        super(Authn, self).__init__(system)
        self._meta_data['allowed_lazy_attributes'] = [
            Providers
        ]


class Providers(OrganizingCollection):
    def __init__(self, authn):
        super(Providers, self).__init__(authn)
        self._meta_data['allowed_lazy_attributes'] = [
            Tmos_s
        ]


class Tmos_s(Collection):
    def __init__(self, providers):
        super(Tmos_s, self).__init__(providers)
        self._meta_data['required_json_kind'] = 'cm:system:authn:providers:tmos:mcpremoteprovidercollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Tmos]
        self._meta_data['attribute_registry'] = {
            'cm:system:authn:providers:tmos:mcpremoteproviderstate': Tmos
        }


class Tmos(Resource):
    def __init__(self, tokens):
        super(Tmos, self).__init__(tokens)
        self._meta_data['required_json_kind'] = 'cm:system:authn:providers:tmos:mcpremoteproviderstate'
        self._meta_data['required_load_parameters'] = set(('id',))

    def create(self, **kwargs):
        raise UnsupportedMethod(
            "Only a single remote mcp auth provider allowed."
        )

    def update(self, **kwargs):
        # You can technically do this, but it will really screw up the system
        raise UnsupportedMethod

    def modify(self, **kwargs):
        # You can technically do this, but it will really screw up the system
        raise UnsupportedMethod

    def delete(self, **kwargs):
        # You can technically do this, but it will really screw up the system
        raise UnsupportedMethod



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Guest (vcmp) module

REST URI
    ``http://localhost/mgmt/tm/vcmp/guest/``

GUI Path
    ``Guest List``

REST Kind
    ``tm:vcmp:guest:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import DisallowedCreationParameter
from f5.sdk_exception import DisallowedReadParameter


class Guests(Collection):
    """BIG-IP® Guests collection."""
    def __init__(self, vcmp):
        super(Guests, self).__init__(vcmp)
        self._meta_data['allowed_lazy_attributes'] = [Guest]
        self._meta_data['required_json_kind'] =\
            'tm:vcmp:guest:guestcollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:vcmp:guest:gueststate': Guest}


class Guest(Resource):
    """BIG-IP® Guest resource."""
    def __init__(self, guests):
        super(Guest, self).__init__(guests)
        self._meta_data['required_json_kind'] =\
            'tm:vcmp:guest:gueststate'

    def _check_load_parameters(self, **kwargs):
        """Override method for one in resource.py to check partition

        The partition cannot be included as a parameter to load a guest.
        Raise an exception if a consumer gives the partition parameter.

        :raises: DisallowedReadParameter
        """

        if 'partition' in kwargs:
            msg = "'partition' is not allowed as a load parameter. Vcmp " \
                "guests are accessed by name."
            raise DisallowedReadParameter(msg)
        super(Guest, self)._check_load_parameters(**kwargs)

    def _check_create_parameters(self, **kwargs):
        """Override method for one in resource.py to check partition

        The partition cannot be included as a parameter to create a guest.
        Raise an exception if a consumer gives the partition parameter.

        :raises: DisallowedCreationParameter
        """

        if 'partition' in kwargs:
            msg = "'partition' is not allowed as a create parameter. Vcmp " \
                "guests are created with the 'name' at least."
            raise DisallowedCreationParameter(msg)
        super(Guest, self)._check_create_parameters(**kwargs)



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® db module

REST URI
    ``http://localhost/mgmt/sys/db/``

GUI Path
    N/A

REST Kind
    ``tm:sys:db:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedOperation


class Dbs(Collection):
    """BIG-IP® db collection"""
    def __init__(self, sys):
        super(Dbs, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [Db]
        self._meta_data['attribute_registry'] =\
            {'tm:sys:db:dbstate': Db}


class Db(Resource):
    """BIG-IP® db resource

    .. note::
        db objects are read-only.
    """
    def __init__(self, dbs):
        super(Db, self).__init__(dbs)
        self._meta_data['required_json_kind'] = 'tm:sys:db:dbstate'

    def create(self, **kwargs):
        '''Create is not supported for db resources.

        :raises: UnsupportedOperation
        '''
        raise UnsupportedOperation(
            "DB resources doesn't support create, only load and refresh"
        )

    def delete(self, **kwargs):
        '''Delete is not supported for db resources.

        :raises: UnsupportedOperation
        '''
        raise UnsupportedOperation(
            "DB resources doesn't support delete, only load and refresh"
        )



# coding=utf-8
# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
"""A base exception for all exceptions in this library."""


class F5SDKError(Exception):
    """Import and subclass this exception in all exceptions in this library."""
    def __init__(self, *args, **kwargs):
        super(F5SDKError, self).__init__(*args, **kwargs)


class AttemptedMutationOfReadOnly(F5SDKError):
    """Read only parameters cannot be set."""
    pass


class BooleansToReduceHaveSameValue(F5SDKError):
    """Dict contains two keys with same boolean value."""
    pass


class DeviceProvidesIncompatibleKey(F5SDKError):
    """Raise this when server JSON keys are incompatible with Python."""
    pass


class DisallowedCreationParameter(F5SDKError):
    """Exception when partition is passed to create for guest resource."""
    pass


class DisallowedReadParameter(F5SDKError):
    """Exception when partition is passed to load for guest resource."""
    pass


class EmptyContent(F5SDKError):
    """Raise an error if the returned content size is 0."""
    pass


class ExclusiveAttributesPresent(F5SDKError):
    """Raises this when exclusive attributes are present."""
    pass


class FileMustNotHaveDotISOExtension(F5SDKError):
    """Raise this when file has ISO extension."""
    def __init__(self, filename):
        super(FileMustNotHaveDotISOExtension, self).__init__(filename)


class GenerationMismatch(F5SDKError):
    """The server reported BIG-IP® is not the expacted value."""
    pass


class ImageFilesMustHaveDotISOExtension(F5SDKError):
    """Raise this when Image files do not have ISO extensions."""
    def __init__(self, filename):
        super(ImageFilesMustHaveDotISOExtension, self).__init__(filename)


class InvalidCommand(F5SDKError):
    """Raise this if command argument supplied is invalid."""
    pass


class InvalidForceType(ValueError):
    """Must be of type bool."""
    pass


class InvalidName(ValueError):
    """Raised during creation when a given resource name is invalid."""
    pass


class InvalidResource(F5SDKError):
    """Raise this when a caller tries to invoke an unsupported CRUDL op.

    All resources support `refresh` and `raw`.
    Only `Resource`'s support `load`, `create`, `update`, and `delete`.
    """
    pass


class KindTypeMismatch(F5SDKError):
    """Raise this when server JSON keys are incorrect for the Resource type."""
    pass


class LazyAttributesRequired(F5SDKError):
    """Raised when a object accesses a lazy attribute that is not listed."""
    pass


class MemberStateModifyUnsupported(F5SDKError):
    """Modify of node with state=unchecked is unsupported."""
    pass


class MissingHttpHeader(F5SDKError):
    """We raise this when the expected http header in response is missing."""
    pass


class MissingRequiredCreationParameter(F5SDKError):
    """Various values MUST be provided to create different Resources."""
    pass


class MissingRequiredCommandParameter(F5SDKError):
    """Various values MUST be provided to execute a command."""
    pass


class MissingRequiredReadParameter(F5SDKError):
    """Various values MUST be provided to refresh some Resources."""
    pass


class MissingUpdateParameter(F5SDKError):
    """Raises this when update requires specific parameters together."""
    pass


class NodeStateModifyUnsupported(F5SDKError):
    """Modify of node with state=unchecked is unsupported."""
    pass


class NonExtantApplication(F5SDKError):
    """Raise if the dos profile application sub-collection

    resource does not exist on the device.
    """
    pass


class NonExtantPolicyRule(F5SDKError):
    """Raise if a rule does not exist on the device."""
    pass


class NonExtantFirewallRule(F5SDKError):
    """Raise if the policy does not exist on the device."""
    pass


class NonExtantVirtualPolicy(F5SDKError):
    """Raise if the policy does not exist on the device."""
    pass


class OperationNotSupportedOnPublishedPolicy(F5SDKError):
    """Raise if update/modify attempted on published policy."""
    pass


class RequestParamKwargCollision(F5SDKError):
    """Raise where requests parameter collides

    with a method parameter of the same name.
    """
    pass


class TagModeDisallowedForTMOSVersion(F5SDKError):
    """Raise if tagmode is not supported for given TMOS version."""
    pass


class TransactionSubmitException(F5SDKError):
    """Raise this when Transaction commit fails."""
    pass


class URICreationCollision(F5SDKError):
    """self._meta_data['uri'] can only be assigned once. In create or load."""
    pass


class UnregisteredKind(F5SDKError):
    """The returned server JSON `kind` key wasn't expected by this Resource."""
    pass


class UnsupportedMethod(F5SDKError):
    """Raise this if a method supplied is unsupported."""
    pass


class UnsupportedTmosVersion(F5SDKError):
    """Raise the error if a class of an API is instantiated,

    on a TMOS version where API was not yet implemented/supported.
    """
    pass


class UnsupportedOperation(F5SDKError):
    """Object does not support the method that was called."""
    pass


class UtilError(F5SDKError):
    """Raise this if command excecution returns an error."""
    pass


class DraftPolicyNotSupportedInTMOSVersion(F5SDKError):
    """Raise when using Drafts in a legacy TMOS version

    Raise this if handling Draft work in a Policy class that is
    used by legacy, and current, versions of BIG-IP
    """
    pass


class ConstraintError(F5SDKError):
    """Raise when a supplied value is outside the limits for that attribute."""
    pass


class RequiredOneOf(F5SDKError):
    """Raise this if more than one of required argument sets is provided."""
    def __init__(self, required_one_of):
        errors = []
        message = 'Creation requires one of the following lists of args {0}'
        for require in required_one_of:
            requires = ','.join(require)
            error = '( {0} )'.format(requires)
            errors.append(error)
        msg = message.format(' or '.join(errors))
        super(RequiredOneOf, self).__init__(msg)


class TimeoutError(F5SDKError):
    """Raised when hard-timeout (timeout keyword to ManagementRoot) is met"""



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""

REST URI
    ``http://localhost/mgmt/tm/shared/license``

REST Kind
    ``tm:shared:licensing:*``
"""

from f5.iworkflow.resource import PathElement
from f5.iworkflow.resource import UnnamedResource
from f5.sdk_exception import UnsupportedMethod


class Licensing(PathElement):
    def __init__(self, shared):
        super(Licensing, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Activation,
            Registration,
        ]
        self._meta_data['attribute_registry'] = {
            'tm:shared:licensing:activation:activatelicenseresponse':
                Activation,
            'tm:shared:licensing:registration:registrationlicenseresponse':
                Registration,
        }


class Activation(UnnamedResource):
    def __init__(self, licensing):
        super(Activation, self).__init__(licensing)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:shared:licensing:activation:activatelicenseresponse'

    def update(self, **kwargs):
        '''Update is not supported for License Activation

        :raises: UnsupportedOperation
        '''
        raise UnsupportedMethod(
            "%s does not support the update method" % self.__class__.__name__
        )


class Registration(UnnamedResource):
    def __init__(self, licensing):
        super(Registration, self).__init__(licensing)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] =\
            'tm:shared:licensing:activation:activatelicenseresponse'



# coding=utf-8
#
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IQ® license pool regkeys.

REST URI
    ``http://localhost/mgmt/cm/device/licensing/pool/initial-activation``

REST Kind
    ``cm:device:licensing:pool:initial-activation:*``
"""

from f5.bigiq.resource import Collection
from f5.bigiq.resource import Resource


class Initial_Activations(Collection):
    def __init__(self, pool):
        super(Initial_Activations, self).__init__(pool)
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:initial-activation:initialactivationworkercollectionstate'  # NOQA
        self._meta_data['allowed_lazy_attributes'] = [Initial_Activation]
        self._meta_data['attribute_registry'] = {
            'cm:device:licensing:pool:initial-activation:initialactivationworkeritemstate': Initial_Activation  # NOQA
        }


class Initial_Activation(Resource):
    def __init__(self, initial_activations):
        super(Initial_Activation, self).__init__(initial_activations)
        self._meta_data['required_creation_parameters'] = {'name', 'regKey'}
        self._meta_data['required_json_kind'] = \
            'cm:device:licensing:pool:initial-activation:initialactivationworkeritemstate'



#!/usr/bin/env python

# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import f5
import os

from setuptools import find_packages
from setuptools import setup

if 'rpm' not in os.getcwd():
    with open('setup_requirements.txt') as fh:
        required = [x for x in fh.read().splitlines() if not x.startswith('#')]
else:
    required = []

setup(
    name='f5-sdk',
    description='F5 Networks Python SDK',
    license='Apache License, Version 2.0',
    version=f5.__version__,
    author='F5 Networks',
    author_email='f5_common_python@f5.com',
    url='https://github.com/F5Networks/f5-common-python',
    keywords=['F5', 'sdk', 'api', 'icontrol', 'bigip', 'api', 'ltm'],
    install_requires=required,
    packages=find_packages(
        exclude=["*.test", "*.test.*", "test.*", "test_*", "test", "test*"]
    ),
    data_files=[('.', ['setup_requirements.txt'])],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Intended Audience :: System Administrators',
    ],
    entry_points={'pytest11': ['f5sdk_fixtures = f5sdk_plugins.fixtures']}
)



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager™ (LTM®) module.

REST URI
    ``http://localhost/mgmt/tm/ltm/``

GUI Path
    ``Local Traffic``

REST Kind
    ``tm:ltm:*``
"""


from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.ltm.auth import Auth
from f5.bigip.tm.ltm.data_group import Data_Group
from f5.bigip.tm.ltm.default_node_monitor import Default_Node_Monitor
from f5.bigip.tm.ltm.ifile import Ifiles
from f5.bigip.tm.ltm.lsn_pools import LSN_Log_Profiles
from f5.bigip.tm.ltm.lsn_pools import LSN_Pools
from f5.bigip.tm.ltm.monitor import Monitor
from f5.bigip.tm.ltm.nat import Nats
from f5.bigip.tm.ltm.node import Nodes
from f5.bigip.tm.ltm.persistence import Persistence
from f5.bigip.tm.ltm.policy import Policys
from f5.bigip.tm.ltm.pool import Pools
from f5.bigip.tm.ltm.profile import Profile
from f5.bigip.tm.ltm.rule import Rules
from f5.bigip.tm.ltm.snat import Snats
from f5.bigip.tm.ltm.snat_translation import Snat_Translations
from f5.bigip.tm.ltm.snatpool import Snatpools
from f5.bigip.tm.ltm.traffic_class import Traffic_Class_s
from f5.bigip.tm.ltm.virtual import Virtuals
from f5.bigip.tm.ltm.virtual_address import Virtual_Address_s


class Ltm(OrganizingCollection):
    """BIG-IP® Local Traffic Manager (LTM) organizing collection."""
    def __init__(self, tm):
        super(Ltm, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Auth,
            Data_Group,
            Default_Node_Monitor,
            Ifiles,
            LSN_Log_Profiles,
            LSN_Pools,
            Monitor,
            Nats,
            Nodes,
            Persistence,
            Policys,
            Pools,
            Profile,
            Rules,
            Snats,
            Snatpools,
            Snat_Translations,
            Traffic_Class_s,
            Virtuals,
            Virtual_Address_s,
        ]



# coding=utf-8
#
# Copyright 2014 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager (LTM) virtual module.

REST URI
    ``http://localhost/mgmt/tm/ltm/virtual``

GUI Path
    ``Local Traffic --> Virtual Servers``

REST Kind
    ``tm:ltm:virtual:*``
"""

from distutils.version import LooseVersion

from f5.bigip.mixins import CheckExistenceMixin
from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import NonExtantVirtualPolicy
from f5.sdk_exception import UnregisteredKind
from f5.sdk_exception import URICreationCollision

from requests import HTTPError


class Virtuals(Collection):
    """BIG-IP® LTM virtual collection"""
    def __init__(self, ltm):
        super(Virtuals, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Virtual]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual:virtualstate': Virtual}


class Virtual(Resource, ExclusiveAttributesMixin):
    """BIG-IP® LTM virtual resource"""
    def __init__(self, virtual_s):
        super(Virtual, self).__init__(virtual_s)
        self._meta_data['allowed_lazy_attributes'] = [Profiles_s]
        self._meta_data['required_json_kind'] = 'tm:ltm:virtual:virtualstate'
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual:profiles:profilescollectionstate': Profiles_s,
             'tm:ltm:virtual:policies:policiescollectionstate': Policies_s}
        self._meta_data['exclusive_attributes'].append(('enabled', 'disabled'))
        self._meta_data['exclusive_attributes'].append(('vlansEnabled', 'vlansDisabled'))

    def load(self, **kwargs):
        result = self._load(**kwargs)
        if not hasattr(result, 'rules'):
            result.__dict__.update({'rules': []})
        return result

    def create(self, **kwargs):
        result = self._create(**kwargs)
        if not hasattr(result, 'rules'):
            result.__dict__.update({'rules': []})
        return result

    def update(self, **kwargs):
        result = self._update(**kwargs)
        if not hasattr(result, 'rules'):
            self.__dict__.update({'rules': []})

    def modify(self, **kwargs):
        result = self._modify(**kwargs)
        if not hasattr(result, 'rules'):
            self.__dict__.update({'rules': []})


class Profiles(Resource, CheckExistenceMixin):
    """BIG-IP® LTM profile resource"""
    def __init__(self, Profiles_s):
        super(Profiles, self).__init__(Profiles_s)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            "tm:ltm:virtual:profiles:profilesstate"

    def _exists(self, **kwargs):
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) == LooseVersion('11.6.0'):
            return self._check_existence_by_collection(self._meta_data['container'], kwargs['name'])
        return super(Profiles, self)._exists(**kwargs)


class Profiles_s(Collection):
    """BIG-IP® LTM profile collection"""
    def __init__(self, virtual):
        super(Profiles_s, self).__init__(virtual)
        self._meta_data['allowed_lazy_attributes'] = [Profiles]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual:profiles:profilesstate': Profiles}


class Policies(Resource, CheckExistenceMixin):
    """BIG-IP® LTM Policies resource"""
    def __init__(self, Policies_s):
        super(Policies, self).__init__(Policies_s)
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['required_json_kind'] =\
            "tm:ltm:virtual:policies:policiesstate"

    def exists(self, **kwargs):
        """check existence of policy under virtual."""
        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])

    def load(self, **kwargs):
        """Override load to retrieve object based on exists above."""
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if self._check_existence_by_collection(
                self._meta_data['container'], kwargs['name']):
            if LooseVersion(tmos_v) == LooseVersion('11.5.4'):
                return self._load_11_5_4(**kwargs)
            else:
                return self._load(**kwargs)
        msg = 'The Policy named, {}, does not exist on the device.'.format(
            kwargs['name'])
        raise NonExtantVirtualPolicy(msg)

    def _load_11_5_4(self, **kwargs):
        """Custom _load method to accommodate for issue in 11.5.4,

        where an existing object would return 404 HTTP response.
        """
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this " \
                    "resource, the _meta_data['uri'] is %s and it should" \
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        refresh_session = self._meta_data['bigip']._meta_data[
            'icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        kwargs.update(requests_params)
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        kwargs = self._check_for_python_keywords(kwargs)
        try:
            response = refresh_session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code != 404:
                raise
            if err.response.status_code == 404:
                return self._return_object(self._meta_data['container'],
                                           kwargs['name'])
        # Make new instance of self
        return self._produce_instance(response)

    def create(self, **kwargs):
        """Custom _create method to accommodate for issue 11.5.4 and 12.1.1,

        Where creation of an object would return 404, despite the object
        being created.
        """
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) == LooseVersion('11.5.4') or LooseVersion(
                tmos_v) == LooseVersion('12.1.1'):
            if 'uri' in self._meta_data:
                error = "There was an attempt to assign a new uri to this " \
                        "resource, the _meta_data['uri'] is %s and it should" \
                        " not be changed." % (self._meta_data['uri'])
                raise URICreationCollision(error)
            self._check_exclusive_parameters(**kwargs)
            requests_params = self._handle_requests_params(kwargs)
            self._check_create_parameters(**kwargs)

            # Reduce boolean pairs as specified by the meta_data entry below
            for key1, key2 in self._meta_data['reduction_forcing_pairs']:
                kwargs = self._reduce_boolean_pair(kwargs, key1, key2)

            # Make convenience variable with short names for this method.
            _create_uri = self._meta_data['container']._meta_data['uri']
            session = self._meta_data['bigip']._meta_data['icr_session']
            # We using try/except just in case some HF will fix
            # this in 11.5.4

            try:
                response = session.post(
                    _create_uri, json=kwargs, **requests_params)

            except HTTPError as err:
                if err.response.status_code != 404:
                    raise
                if err.response.status_code == 404:
                    return self._return_object(self._meta_data['container'],
                                               kwargs['name'])
            # Make new instance of self
            return self._produce_instance(response)
        else:
            return self._create(**kwargs)


class Policies_s(Collection):
    """BIG-IP® LTM Policies resource"""
    def __init__(self, virtual):
        super(Policies_s, self).__init__(virtual)
        self._meta_data['allowed_lazy_attributes'] = [Policies]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual:policies:policiesstate': Policies}

    def get_collection(self, **kwargs):
        """We need special get collection method to address issue in 11.5.4

        In 11.5.4 collection 'items' were nested under 'policiesReference'
        key. This has caused get_collection() calls to return empty list.
        This fix will update the list if the policiesReference key is found
        and 'items' key do not exists in __dict__.

        :raises: UnregisteredKind
        :returns: list of reference dicts and Python ``Resource`` objects
        """
        list_of_contents = []
        self.refresh(**kwargs)
        if 'items' in self.__dict__:
            for item in self.items:
                # It's possible to have non-"kind" JSON returned. We just
                # append the corresponding dict. PostProcessing is the caller's
                # responsibility.
                if 'kind' not in item:
                    list_of_contents.append(item)
                    continue
                kind = item['kind']
                if kind in self._meta_data['attribute_registry']:
                    # If it has a kind, it must be registered.
                    instance =\
                        self._meta_data['attribute_registry'][kind](self)
                    instance._local_update(item)
                    instance._activate_URI(instance.selfLink)
                    list_of_contents.append(instance)
                else:
                    error_message = '%r is not registered!' % kind
                    raise UnregisteredKind(error_message)

        if 'policiesReference' in self.__dict__ and 'items' not in \
                self.__dict__:
            for item in self.policiesReference['items']:
                kind = item['kind']
                if kind in self._meta_data['attribute_registry']:
                    # If it has a kind, it must be registered.
                    instance = \
                        self._meta_data['attribute_registry'][kind](self)
                    instance._local_update(item)
                    instance._activate_URI(instance.selfLink)
                    list_of_contents.append(instance)
                else:
                    error_message = '%r is not registered!' % kind
                    raise UnregisteredKind(error_message)

        return list_of_contents



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IQ® License (shared) module

REST URI
    ``http://localhost/mgmt/cm/shared/licensing/``

REST Kind
    N/A -- HTTP GET returns an error
"""

from f5.bigiq.cm.shared.licensing.pools import Pools_s
from f5.bigiq.resource import OrganizingCollection


class Licensing(OrganizingCollection):
    def __init__(self, shared):
        super(Licensing, self).__init__(shared)
        self._meta_data['allowed_lazy_attributes'] = [
            Pools_s
        ]



# coding=utf-8
#
"""Classes and functions for configuring BIG-IP"""
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from f5.bigip.mixins import FileDownloadMixin
from f5.bigip.mixins import FileUploadMixin
from f5.bigip.resource import PathElement
from f5.sdk_exception import ImageFilesMustHaveDotISOExtension


class Software_Image_Uploads(PathElement, FileUploadMixin):
    """Software image upload resource."""
    def __init__(self, autodeploy):
        super(Software_Image_Uploads, self).__init__(autodeploy)
        self._meta_data['minimum_version'] = '12.0.0'

    def upload_image(self, filepathname, **kwargs):
        filename = os.path.basename(filepathname)
        if os.path.splitext(filename)[-1] != '.iso':
            raise ImageFilesMustHaveDotISOExtension(filename)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._upload_file(filepathname, **kwargs)


class Software_Image_Downloads(PathElement, FileDownloadMixin):
    """Software image download resource."""
    def __init__(self, autodeploy):
        super(Software_Image_Downloads, self).__init__(autodeploy)
        self._meta_data['minimum_version'] = '12.0.0'

    def download_image(self, src, dest, **kwargs):
        filename = os.path.basename(src)
        self.file_bound_uri = self._meta_data['uri'] + filename
        self._download_file(filename, dest, **kwargs)

    def download_file(self, src, dest, **kwargs):
        self.download_image(src, dest, **kwargs)



# coding=utf-8
#
#  Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®) Region module.

REST URI
    ``http://localhost/mgmt/tm/gtm/region``

GUI Path
    ``DNS --> GSLB : Topology : Regions``

REST Kind
    ``tm:gtm:region:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Regions(Collection):
    """BIG-IP® GTM Region collection"""
    def __init__(self, gtm):
        super(Regions, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Region]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:region:regionstate': Region}


class Region(Resource):
    """BIG-IP® GTM Region resource"""
    def __init__(self, regions):
        super(Region, self).__init__(regions)
        self._meta_data['required_json_kind'] = 'tm:gtm:region:regionstate'
